@@ -184,13 +184,7 @@ static AcSecMap* alloc_secondary_map ( __attribute__ ((unused))
    PROF_EVENT(10);
 
    /* Mark all bytes as invalid access and invalid value. */
-
-   /* It just happens that a AcSecMap occupies exactly 18 pages --
-      although this isn't important, so the following assert is
-      spurious. */
-   sk_assert(0 == (sizeof(AcSecMap) % VKI_BYTES_PER_PAGE));
    map = (AcSecMap *)VG_(shadow_alloc)(sizeof(AcSecMap));
-
    for (i = 0; i < 8192; i++)
       map->abits[i] = VGM_BYTE_INVALID; /* Invalid address */
 
@@ -721,34 +715,34 @@ static __inline__ void ac_helperc_ACCESS1 ( Addr a, Bool isWrite )
 #  endif
 }
 
-__attribute__ ((regparm(1)))
+REGPARM(1)
 static void ac_helperc_LOAD4 ( Addr a )
 {
    ac_helperc_ACCESS4 ( a, /*isWrite*/False );
 }
-__attribute__ ((regparm(1)))
+REGPARM(1)
 static void ac_helperc_STORE4 ( Addr a )
 {
    ac_helperc_ACCESS4 ( a, /*isWrite*/True );
 }
 
-__attribute__ ((regparm(1)))
+REGPARM(1)
 static void ac_helperc_LOAD2 ( Addr a )
 {
    ac_helperc_ACCESS2 ( a, /*isWrite*/False );
 }
-__attribute__ ((regparm(1)))
+REGPARM(1)
 static void ac_helperc_STORE2 ( Addr a )
 {
    ac_helperc_ACCESS2 ( a, /*isWrite*/True );
 }
 
-__attribute__ ((regparm(1)))
+REGPARM(1)
 static void ac_helperc_LOAD1 ( Addr a )
 {
    ac_helperc_ACCESS1 ( a, /*isWrite*/False );
 }
-__attribute__ ((regparm(1)))
+REGPARM(1)
 static void ac_helperc_STORE1 ( Addr a )
 {
    ac_helperc_ACCESS1 ( a, /*isWrite*/True );
@@ -918,13 +912,13 @@ void ac_fpu_ACCESS_check ( Addr addr, Int size, Bool isWrite )
 #  endif
 }
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 static void ac_fpu_READ_check ( Addr addr, Int size )
 {
    ac_fpu_ACCESS_check ( addr, size, /*isWrite*/False );
 }
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 static void ac_fpu_WRITE_check ( Addr addr, Int size )
 {
    ac_fpu_ACCESS_check ( addr, size, /*isWrite*/True );
@@ -1046,6 +1040,7 @@ UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
             VG_(copy_UInstr)(cb, u_in);
             break;
 
+         case MMX2a1_MemRd:
          case SSE3a_MemRd:
          case SSE2a_MemRd:
          case SSE3a1_MemRd:
@@ -1125,9 +1120,9 @@ Bool ac_is_valid_address ( Addr a )
 }
 
 
-/* Leak detector for this skin.  We don't actually do anything, merely
+/* Leak detector for this tool.  We don't actually do anything, merely
    run the generic leak detector with suitable parameters for this
-   skin. */
+   tool. */
 static void ac_detect_memory_leaks ( void )
 {
    MAC_(do_detect_memory_leaks) ( ac_is_valid_64k_chunk, ac_is_valid_address );
@@ -1184,7 +1179,11 @@ Bool SK_(handle_client_request) ( ThreadId tid, UInt* arg_block, UInt *ret )
    /* Overload memcheck client reqs */
    if (!VG_IS_SKIN_USERREQ('M','C',arg[0])
     && VG_USERREQ__MALLOCLIKE_BLOCK != arg[0]
-    && VG_USERREQ__FREELIKE_BLOCK   != arg[0])
+    && VG_USERREQ__FREELIKE_BLOCK   != arg[0]
+    && VG_USERREQ__CREATE_MEMPOOL   != arg[0]
+    && VG_USERREQ__DESTROY_MEMPOOL  != arg[0]
+    && VG_USERREQ__MEMPOOL_ALLOC    != arg[0]
+    && VG_USERREQ__MEMPOOL_FREE     != arg[0])
       return False;
 
    switch (arg[0]) {
@@ -1258,7 +1257,7 @@ void SK_(pre_clo_init)(void)
    VG_(details_version)         (NULL);
    VG_(details_description)     ("a fine-grained address checker");
    VG_(details_copyright_author)(
-      "Copyright (C) 2002-2004, and GNU GPL'd, by Julian Seward.");
+      "Copyright (C) 2002-2004, and GNU GPL'd, by Julian Seward et al.");
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
    VG_(details_avg_translation_sizeB) ( 135 );
 
