@@ -6,7 +6,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward
+   Copyright (C) 2000-2004 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -215,7 +215,7 @@ int process_extended_line_op( SegInfo *si, Char*** fnames,
         *fnames = VG_(arena_malloc)(VG_AR_SYMTAB, sizeof (UInt) * 2);
       else
         *fnames = VG_(arena_realloc)(
-                     VG_AR_SYMTAB, *fnames, /*alignment*/4,
+                     VG_AR_SYMTAB, *fnames, VG_MIN_MALLOC_SZB,
                      sizeof(UInt) 
                         * (state_machine_regs.last_file_entry + 1));
       (*fnames)[state_machine_regs.last_file_entry] = VG_(addStr) (si,name, -1);
@@ -367,7 +367,8 @@ void VG_(read_debuginfo_dwarf2) ( SegInfo* si, UChar* dwarf2, Int dwarf2_sz )
              if (fnames == NULL)
                fnames = VG_(arena_malloc)(VG_AR_SYMTAB, sizeof (UInt) * 2);
              else
-               fnames = VG_(arena_realloc)(VG_AR_SYMTAB, fnames, /*alignment*/4,
+               fnames = VG_(arena_realloc)(VG_AR_SYMTAB, fnames,
+                           VG_MIN_MALLOC_SZB,
                            sizeof(UInt) 
                               * (state_machine_regs.last_file_entry + 1));
              data += VG_(strlen) ((Char *) data) + 1;
