@@ -43,15 +43,153 @@ static Error* vg_errors = NULL;
 static Supp* vg_suppressions = NULL;
 
 /* Running count of unsuppressed errors detected. */
-UInt VG_(n_errs_found) = 0;
+static UInt n_errs_found = 0;
 
 /* Running count of suppressed errors detected. */
-static UInt vg_n_errs_suppressed = 0;
+static UInt n_errs_suppressed = 0;
 
 /* forwards ... */
 static Supp* is_suppressible_error ( Error* err );
 
 
+/*------------------------------------------------------------*/
+/*--- Error type                                           ---*/
+/*------------------------------------------------------------*/
+
+/* Note: it is imperative this doesn't overlap with (0..) at all, as tools
+ * effectively extend it by defining their own enums in the (0..) range. */
+typedef
+   enum { 
+      PThreadErr      = -1,   // Pthreading error
+   }
+   CoreErrorKind;
+
+/* Errors.  Extensible (via the 'extra' field).  Tools can use a normal
+   enum (with element values in the normal range (0..)) for `ekind'. 
+   Functions for getting/setting the tool-relevant fields are in
+   include/vg_skin.h.
+
+   When errors are found and recorded with VG_(maybe_record_error)(), all
+   the tool must do is pass in the four parameters;  core will
+   allocate/initialise the error record.
+*/
+struct _Error {
+   struct _Error* next;
+   // NULL if unsuppressed; or ptr to suppression record.
+   Supp* supp;
+   Int count;
+   ThreadId tid;
+
+   // The tool-specific part
+   ExeContext* where;      // Initialised by core
+   Int ekind;              // Used by ALL.  Must be in the range (0..)
+   Addr addr;              // Used frequently
+   Char* string;           // Used frequently
+   void* extra;            // For any tool-specific extras
+};
+
+ExeContext* VG_(get_error_where) ( Error* err )
+{
+   return err->where;
+}
+
+ErrorKind VG_(get_error_kind) ( Error* err )
+{
+   return err->ekind;
+}
+
+Addr VG_(get_error_address) ( Error* err )
+{
+   return err->addr;
+}
+
+Char* VG_(get_error_string) ( Error* err )
+{
+   return err->string;
+}
+
+void* VG_(get_error_extra)  ( Error* err )
+{
+   return err->extra;
+}
+
+UInt VG_(get_n_errs_found)( void )
+{
+   return n_errs_found;
+}
+
+/*------------------------------------------------------------*/
+/*--- Suppression type                                     ---*/
+/*------------------------------------------------------------*/
+
+/* Note: it is imperative this doesn't overlap with (0..) at all, as tools
+ * effectively extend it by defining their own enums in the (0..) range. */
+typedef
+   enum {
+      PThreadSupp = -1,    /* Matches PThreadErr */
+   }
+   CoreSuppKind;
+
+/* For each caller specified for a suppression, record the nature of
+   the caller name.  Not of interest to tools. */
+typedef
+   enum { 
+      ObjName,    /* Name is of an shared object file. */
+      FunName     /* Name is of a function. */
+   }
+   SuppLocTy;
+
+/* Suppressions.  Tools can get/set tool-relevant parts with functions
+   declared in include/vg_skin.h.  Extensible via the 'extra' field. 
+   Tools can use a normal enum (with element values in the normal range
+   (0..)) for `skind'. */
+struct _Supp {
+   struct _Supp* next;
+   Int count;     // The number of times this error has been suppressed.
+   Char* sname;   // The name by which the suppression is referred to.
+   /* First two (name of fn where err occurs, and immediate caller)
+    * are mandatory;  extra two are optional. */
+   SuppLocTy caller_ty[VG_N_SUPP_CALLERS];
+   Char*     caller   [VG_N_SUPP_CALLERS];
+
+   /* The tool-specific part */
+   SuppKind skind;   // What kind of suppression.  Must use the range (0..).
+   Char* string;     // String -- use is optional.  NULL by default.
+   void* extra;      // Anything else -- use is optional.  NULL by default.
+};
+
+SuppKind VG_(get_supp_kind) ( Supp* su )
+{
+   return su->skind;
+}
+
+Char* VG_(get_supp_string) ( Supp* su )
+{
+   return su->string;
+}
+
+void* VG_(get_supp_extra)  ( Supp* su )
+{
+   return su->extra;
+}
+
+
+void VG_(set_supp_kind)   ( Supp* su, SuppKind skind )
+{
+   su->skind = skind;
+}
+
+void VG_(set_supp_string) ( Supp* su, Char* string )
+{
+   su->string = string;
+}
+
+void VG_(set_supp_extra)  ( Supp* su, void* extra )
+{
+   su->extra = extra;
+}
+
+
 /*------------------------------------------------------------*/
 /*--- Helper fns                                           ---*/
 /*------------------------------------------------------------*/
@@ -181,7 +319,7 @@ void construct_error ( Error* err, ThreadId tid, ErrorKind ekind, Addr a,
    else
       err->where = where;
 
-   /* Skin-relevant parts */
+   /* Tool-relevant parts */
    err->ekind  = ekind;
    err->addr   = a;
    err->extra  = extra;
@@ -238,7 +376,7 @@ static void gen_suppression(Error* err)
          VG_(printf)("   obj:%s\n", buf);
       } else {
          VG_(printf)("   ???:???       "
-                     "# unknown, suppression will not work, sorry)\n");
+                     "# unknown, suppression will not work, sorry\n");
       }
       i++;
    } while (i < stop_at && ec->eips[i] != 0);
@@ -290,7 +428,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
       pointless to continue the Valgrind run after this point. */
    if (VG_(clo_error_limit) 
        && (vg_n_errs_shown >= M_VG_COLLECT_NO_ERRORS_AFTER_SHOWN
-           || VG_(n_errs_found) >= M_VG_COLLECT_NO_ERRORS_AFTER_FOUND)) {
+           || n_errs_found >= M_VG_COLLECT_NO_ERRORS_AFTER_FOUND)) {
       if (!stopping_message) {
          VG_(message)(Vg_UserMsg, "");
 
@@ -349,9 +487,9 @@ void VG_(maybe_record_error) ( ThreadId tid,
 	 if (p->supp != NULL) {
             /* Deal correctly with suppressed errors. */
             p->supp->count++;
-            vg_n_errs_suppressed++;	 
+            n_errs_suppressed++;	 
          } else {
-            VG_(n_errs_found)++;
+            n_errs_found++;
          }
 
          /* Move p to the front of the list so that future searches
@@ -406,7 +544,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
    p->supp = is_suppressible_error(&err);
    vg_errors = p;
    if (p->supp == NULL) {
-      VG_(n_errs_found)++;
+      n_errs_found++;
       if (!is_first_shown_context)
          VG_(message)(Vg_UserMsg, "");
       pp_Error(p, False);
@@ -414,7 +552,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
       vg_n_errs_shown++;
       do_actions_on_error(p, /*allow_db_attach*/True);
    } else {
-      vg_n_errs_suppressed++;
+      n_errs_suppressed++;
       p->supp->count++;
    }
 }
@@ -424,7 +562,7 @@ void VG_(maybe_record_error) ( ThreadId tid,
    guaranteed to only happen once.  This avoids all the recording and
    comparing stuff.  But they can be suppressed;  returns True if it is
    suppressed.  Bool `print_error' dictates whether to print the error. 
-   Bool `count_error' dictates whether to count the error in VG_(n_errs_found)
+   Bool `count_error' dictates whether to count the error in n_errs_found.
 */
 Bool VG_(unique_error) ( ThreadId tid, ErrorKind ekind, Addr a, Char* s,
                          void* extra, ExeContext* where, Bool print_error,
@@ -445,7 +583,7 @@ Bool VG_(unique_error) ( ThreadId tid, ErrorKind ekind, Addr a, Char* s,
 
    if (NULL == is_suppressible_error(&err)) {
       if (count_error)
-         VG_(n_errs_found)++;
+         n_errs_found++;
 
       if (print_error) {
          if (!is_first_shown_context)
@@ -458,7 +596,7 @@ Bool VG_(unique_error) ( ThreadId tid, ErrorKind ekind, Addr a, Char* s,
       return False;
 
    } else {
-      vg_n_errs_suppressed++;
+      n_errs_suppressed++;
       return True;
    }
 }
@@ -476,8 +614,6 @@ void VG_(record_pthread_error) ( ThreadId tid, Char* msg )
    VG_(maybe_record_error)( tid, PThreadErr, /*addr*/0, msg, /*extra*/NULL );
 }
 
-/*------------------------------*/
-
 void VG_(show_all_errors) ( void )
 {
    Int    i, n_min;
@@ -503,8 +639,8 @@ void VG_(show_all_errors) ( void )
    VG_(message)(Vg_UserMsg,
                 "ERROR SUMMARY: "
                 "%d errors from %d contexts (suppressed: %d from %d)",
-                VG_(n_errs_found), n_err_contexts, 
-                vg_n_errs_suppressed, n_supp_contexts );
+                n_errs_found, n_err_contexts, 
+                n_errs_suppressed, n_supp_contexts );
 
    if (VG_(clo_verbosity) <= 1)
       return;
@@ -529,8 +665,8 @@ void VG_(show_all_errors) ( void )
       pp_Error( p_min, False );
 
       if ((i+1 == VG_(clo_dump_error))) {
-	VG_(translate) ( 0 /* dummy ThreadId; irrelevant due to below NULLs */,
-                         p_min->where->eips[0], NULL, NULL, NULL, NULL );
+	VG_(translate) ( 0 /* dummy ThreadId; irrelevant due to debugging*/,
+                         p_min->where->eips[0], /*debugging*/True);
       }
 
       p_min->count = 1 << 30;
@@ -552,8 +688,7 @@ void VG_(show_all_errors) ( void )
       VG_(message)(Vg_UserMsg,
                    "IN SUMMARY: "
                    "%d errors from %d contexts (suppressed: %d from %d)",
-                   VG_(n_errs_found), n_err_contexts, 
-                   vg_n_errs_suppressed,
+                   n_errs_found, n_err_contexts, n_errs_suppressed,
                    n_supp_contexts );
       VG_(message)(Vg_UserMsg, "");
    }
