@@ -681,14 +681,16 @@ static Int proxylwp(void *v)
 	       want while running the handler. */
 	    vg_assert(px->state == PXS_SigACK);
 	    appsigmask = req.sigmask;
-	    VG_(ksigdelset)(&appsigmask, VKI_SIGVGKILL);  /* but allow SIGVGKILL to interrupt */
+	    VG_(ksigdelset)(&appsigmask, VKI_SIGVGKILL);  /* but allow SIGVGKILL */
+	    VG_(ksigdelset)(&appsigmask, VKI_SIGVGINT);   /*       and SIGVGINT to interrupt */
 	    px->state = PXS_WaitReq;
 	    reply.req = PX_BAD;	/* don't reply */
 	    break;
 	    
 	 case PX_SetSigmask:
 	    appsigmask = req.sigmask;
-	    VG_(ksigdelset)(&appsigmask, VKI_SIGVGKILL);   /* but allow SIGVGKILL to interrupt */
+	    VG_(ksigdelset)(&appsigmask, VKI_SIGVGKILL);  /* but allow SIGVGKILL */
+	    VG_(ksigdelset)(&appsigmask, VKI_SIGVGINT);   /*       and SIGVGINT to interrupt */
 
 	    vg_assert(px->state == PXS_WaitReq || 
 		      px->state == PXS_SigACK);
@@ -877,11 +879,6 @@ static Int do_futex(void *addr, Int op, Int val, struct vki_timespec *time, void
    return VG_(do_syscall)(__NR_futex, addr, op, val, time, addr2);
 }
 
-#define VKI_FUTEX_WAIT		0
-#define VKI_FUTEX_WAKE		1
-#define VKI_FUTEX_FD		2
-#define VKI_FUTEX_REQUEUE	3
-
 static Int have_settid = -1;	/* -1 -> unknown */
 
 /*
@@ -1317,7 +1314,7 @@ Int VG_(sys_issue)(int tid)
 }
 
 /* Relatively expensive sanity tests for the syscall machinery */
-void VG_(proxy_sanity)(void)
+void VG_(sanity_check_proxy)(void)
 {
    Int tid;
    Bool sane = True;
@@ -1375,12 +1372,11 @@ void VG_(proxy_sanity)(void)
 }
 
 /* Get the PID/TID of the ProxyLWP. */
-Int VG_(proxy_id)(ThreadId tid)
+__attribute__((unused))
+static Int proxy_id(ThreadId tid)
 {
    ThreadState *tst = VG_(get_ThreadState)(tid);
-   ProxyLWP *proxy = tst->proxy;
-
-   return proxy->lwp;
+   return tst->proxy->lwp;
 }
 
 /*--------------------------------------------------------------------*/
