@@ -64,8 +64,7 @@
    with shared signal state (Linux 2.5+, RedHat's 2.4), then these are
    the only signals it needs to handle.
 
-   If we get a synchronous signal, the details are placed into
-   VG_(unresumable_siginfo) and we longjmp back into the scheduler,
+   If we get a synchronous signal, we longjmp back into the scheduler,
    since we can't resume executing the client code.  The scheduler
    immediately starts signal delivery to the thread which generated
    the signal.
@@ -82,16 +81,6 @@
 #define DEBUG_SIGNALS
 
 
-/*
-   - The following causes an infinite loop: start Hugs, Feb 2001 
-     version, and do Control-C at the prompt.  There is an infinite
-     series of sigints delivered (to the client); but also seemingly
-     to valgrind, which is very strange.  I don't know why.
-
-     [I haven't re-tested this, but this is likely fixed - JSGF]
-*/
-
-
 /* ---------------------------------------------------------------------
    Forwards decls.
    ------------------------------------------------------------------ */
@@ -104,6 +93,24 @@ static void proxy_sigvg_handler	   ( Int sigNo, vki_ksiginfo_t *info, struct vki
 static Bool is_correct_sigmask(void);
 static const Char *signame(Int sigNo);
 
+/* ---------------------------------------------------------------------
+   Signal stack
+   ------------------------------------------------------------------ */
+
+/* Valgrind's signal stack size, in words */
+#define VG_SIGSTACK_SIZE_W    10000
+
+/* We have to ask for signals to be delivered on an alternative
+   stack, since it is possible, although unlikely, that we'll have to run
+   client code from inside the Valgrind-installed signal handler. */
+static Addr sigstack[VG_SIGSTACK_SIZE_W];
+
+extern void VG_(get_sigstack_bounds)( Addr* low, Addr* high )
+{
+   *low  = (Addr) & sigstack[0];
+   *high = (Addr) & sigstack[VG_SIGSTACK_SIZE_W]; 
+}
+
 /* ---------------------------------------------------------------------
    HIGH LEVEL STUFF TO DO WITH SIGNALS: POLICY (MOSTLY)
    ------------------------------------------------------------------ */
@@ -173,9 +180,6 @@ typedef
       /* per-signal info */
       SCSS_Per_Signal scss_per_sig[1+VKI_KNSIG];
 
-      /* Signal delivery stack, if any. */
-      vki_kstack_t altstack;
-
       /* Additional elements to SCSS not stored here:
          - for each thread, the thread's blocking mask
          - for each thread in WaitSIG, the set of waited-on sigs
@@ -355,6 +359,9 @@ void calculate_SKSS_from_SCSS ( SKSS* dst )
       /* always ask for SA_SIGINFO */
       skss_flags |= VKI_SA_SIGINFO;
 
+      /* use our own restorer */
+      skss_flags |= VKI_SA_RESTORER;
+
       /* Create SKSS entry for this signal. */
 
       if (sig != VKI_SIGKILL && sig != VKI_SIGSTOP)
@@ -380,7 +387,7 @@ void calculate_SKSS_from_SCSS ( SKSS* dst )
    After a possible SCSS change, update SKSS and the kernel itself.
    ------------------------------------------------------------------ */
 
-void VG_(handle_SCSS_change) ( Bool force_update )
+static void handle_SCSS_change ( Bool force_update )
 {
    Int            res, sig;
    SKSS           skss_old;
@@ -412,11 +419,12 @@ void VG_(handle_SCSS_change) ( Bool force_update )
 
       ksa.ksa_handler = vg_skss.skss_per_sig[sig].skss_handler;
       ksa.ksa_flags   = vg_skss.skss_per_sig[sig].skss_flags;
+      ksa.ksa_restorer = VG_(sigreturn);
+
       vg_assert(ksa.ksa_flags & VKI_SA_ONSTACK);
       VG_(ksigfillset)( &ksa.ksa_mask );
       VG_(ksigdelset)( &ksa.ksa_mask, VKI_SIGKILL );
       VG_(ksigdelset)( &ksa.ksa_mask, VKI_SIGSTOP );
-      ksa.ksa_restorer = NULL;
 
       if (VG_(clo_trace_signals)) 
          VG_(message)(Vg_DebugMsg, 
@@ -439,7 +447,7 @@ void VG_(handle_SCSS_change) ( Bool force_update )
          vg_assert(ksa_old.ksa_flags 
                    == skss_old.skss_per_sig[sig].skss_flags);
          vg_assert(ksa_old.ksa_restorer 
-                   == NULL);
+                   == VG_(sigreturn));
          VG_(ksigaddset)( &ksa_old.ksa_mask, VKI_SIGKILL );
          VG_(ksigaddset)( &ksa_old.ksa_mask, VKI_SIGSTOP );
          vg_assert(VG_(kisfullsigset)( &ksa_old.ksa_mask ));
@@ -456,17 +464,21 @@ void VG_(handle_SCSS_change) ( Bool force_update )
    in kernel/signal.[ch] */
 
 /* True if we are on the alternate signal stack.  */
-static Int on_sig_stack ( Addr m_esp )
+static Int on_sig_stack ( ThreadId tid, Addr m_esp )
 {
-   return (m_esp - (Addr)vg_scss.altstack.ss_sp 
-           < vg_scss.altstack.ss_size);
+   ThreadState *tst = VG_(get_ThreadState)(tid);
+
+   return (m_esp - (Addr)tst->altstack.ss_sp 
+           < tst->altstack.ss_size);
 }
 
-static Int sas_ss_flags ( Addr m_esp )
+static Int sas_ss_flags ( ThreadId tid, Addr m_esp )
 {
-   return (vg_scss.altstack.ss_size == 0 
+   ThreadState *tst = VG_(get_ThreadState)(tid);
+
+   return (tst->altstack.ss_size == 0 
               ? VKI_SS_DISABLE
-              : on_sig_stack(m_esp) ? VKI_SS_ONSTACK : 0);
+              : on_sig_stack(tid, m_esp) ? VKI_SS_ONSTACK : 0);
 }
 
 
@@ -488,13 +500,13 @@ void VG_(do__NR_sigaltstack) ( ThreadId tid )
          tid, (UInt)ss, (UInt)oss, (UInt)m_esp );
 
    if (oss != NULL) {
-      oss->ss_sp    = vg_scss.altstack.ss_sp;
-      oss->ss_size  = vg_scss.altstack.ss_size;
-      oss->ss_flags = vg_scss.altstack.ss_flags | sas_ss_flags(m_esp);
+      oss->ss_sp    = VG_(threads)[tid].altstack.ss_sp;
+      oss->ss_size  = VG_(threads)[tid].altstack.ss_size;
+      oss->ss_flags = VG_(threads)[tid].altstack.ss_flags | sas_ss_flags(tid, m_esp);
    }
 
    if (ss != NULL) {
-      if (on_sig_stack(VG_(threads)[tid].m_esp)) {
+      if (on_sig_stack(tid, VG_(threads)[tid].m_esp)) {
          SET_SYSCALL_RETVAL(tid, -VKI_EPERM);
          return;
       }
@@ -505,16 +517,16 @@ void VG_(do__NR_sigaltstack) ( ThreadId tid )
          return;
       }
       if (ss->ss_flags == VKI_SS_DISABLE) {
-         vg_scss.altstack.ss_flags = VKI_SS_DISABLE;
+         VG_(threads)[tid].altstack.ss_flags = VKI_SS_DISABLE;
       } else {
          if (ss->ss_size < VKI_MINSIGSTKSZ) {
             SET_SYSCALL_RETVAL(tid, -VKI_ENOMEM);
             return;
          }
 
-	 vg_scss.altstack.ss_sp   = ss->ss_sp;
-	 vg_scss.altstack.ss_size = ss->ss_size;
-	 vg_scss.altstack.ss_flags = 0;
+	 VG_(threads)[tid].altstack.ss_sp   = ss->ss_sp;
+	 VG_(threads)[tid].altstack.ss_size = ss->ss_size;
+	 VG_(threads)[tid].altstack.ss_flags = 0;
       }
    }
    SET_SYSCALL_RETVAL(tid, 0);
@@ -579,7 +591,7 @@ void VG_(do__NR_sigaction) ( ThreadId tid )
 
    /* All happy bunnies ... */
    if (new_act) {
-      VG_(handle_SCSS_change)( False /* lazy update */ );
+      handle_SCSS_change( False /* lazy update */ );
    }
    SET_SYSCALL_RETVAL(tid, 0);
    return;
@@ -726,20 +738,6 @@ void VG_(do_pthread_sigmask_SCSS_upd) ( ThreadId tid,
 }
 
 
-void VG_(send_signal_to_thread) ( ThreadId thread, Int sig )
-{
-   ThreadState *tst;
-
-   vg_assert(VG_(is_valid_tid)(thread));
-   vg_assert(sig >= 1 && sig <= VKI_KNSIG);
-   
-   tst = VG_(get_ThreadState)(thread);
-   vg_assert(tst->proxy != NULL);
-
-   VG_(proxy_sendsig)(thread, sig);
-}
-
-
 /* ---------------------------------------------------------------------
    LOW LEVEL STUFF TO DO WITH SIGNALS: IMPLEMENTATION
    ------------------------------------------------------------------ */
@@ -921,7 +919,7 @@ static void synth_ucontext(ThreadId tid, const vki_ksiginfo_t *si,
    uc->uc_flags = 0;
    uc->uc_link = 0;
    uc->uc_sigmask = *set;
-   uc->uc_stack = vg_scss.altstack;
+   uc->uc_stack = tst->altstack;
 
 #define SC(reg)	sc->reg = tst->m_##reg
    SC(gs);
@@ -975,23 +973,23 @@ void vg_push_signal_frame ( ThreadId tid, const vki_ksiginfo_t *siginfo )
        && /* there is a defined and enabled alt stack, which we're not
              already using.  Logic from get_sigframe in
              arch/i386/kernel/signal.c. */
-          sas_ss_flags(tst->m_esp) == 0
+          sas_ss_flags(tid, tst->m_esp) == 0
       ) {
       esp_top_of_frame 
-         = (Addr)(vg_scss.altstack.ss_sp) + vg_scss.altstack.ss_size;
+         = (Addr)(tst->altstack.ss_sp) + tst->altstack.ss_size;
       if (VG_(clo_trace_signals))
          VG_(message)(Vg_DebugMsg,
             "delivering signal %d (%s) to thread %d: on ALT STACK", 
             sigNo, signame(sigNo), tid );
 
       /* Signal delivery to tools */
-      VG_TRACK( pre_deliver_signal, tid, sigNo, /*alt_stack*/False );
+      VG_TRACK( pre_deliver_signal, tid, sigNo, /*alt_stack*/True );
       
    } else {
       esp_top_of_frame = tst->m_esp;
 
       /* Signal delivery to tools */
-      VG_TRACK( pre_deliver_signal, tid, sigNo, /*alt_stack*/True );
+      VG_TRACK( pre_deliver_signal, tid, sigNo, /*alt_stack*/False );
    }
 
    esp = esp_top_of_frame;
@@ -1026,6 +1024,9 @@ void vg_push_signal_frame ( ThreadId tid, const vki_ksiginfo_t *siginfo )
       VG_TRACK( pre_mem_write, Vg_CoreSignal, tid, "signal handler frame (siginfo)", 
 		(Addr)&frame->sigInfo, sizeof(frame->sigInfo) );
       VG_(memcpy)(&frame->sigInfo, siginfo, sizeof(vki_ksiginfo_t));
+      if (sigNo == VKI_SIGFPE) {
+         frame->sigInfo._sifields._sigfault._addr = (void *)tst->m_eip;
+      }
       VG_TRACK( post_mem_write, (Addr)&frame->sigInfo, sizeof(frame->sigInfo) );
 
       VG_TRACK( pre_mem_write, Vg_CoreSignal, tid, "signal handler frame (siginfo)", 
@@ -1276,6 +1277,394 @@ void VG_(kill_self)(Int sigNo)
    VG_(ksigprocmask)(VKI_SIG_SETMASK, &origmask, NULL);
 }
 
+/*
+  Dump core
+   
+  Generate a standard ELF core file corresponding to the client state
+  at the time of a crash.
+ */
+#include <elf.h>
+#ifndef NT_PRXFPREG
+#define NT_PRXFPREG     0x46e62b7f      /* copied from gdb5.1/include/elf/common.h */
+#endif /* NT_PRXFPREG */
+
+/* If true, then this Segment may be mentioned in the core */
+static Bool may_dump(const Segment *seg)
+{
+   return (seg->flags & SF_VALGRIND) == 0 && VG_(is_client_addr)(seg->addr);
+}
+
+/* If true, then this Segment's contents will be in the core */
+static Bool should_dump(const Segment *seg)
+{
+   return may_dump(seg); // && (seg->prot & VKI_PROT_WRITE);
+}
+
+static void fill_ehdr(Elf32_Ehdr *ehdr, Int num_phdrs)
+{
+   VG_(memset)(ehdr, 0, sizeof(ehdr));
+
+   VG_(memcpy)(ehdr->e_ident, ELFMAG, SELFMAG);
+   ehdr->e_ident[EI_CLASS] = ELFCLASS32;
+   ehdr->e_ident[EI_DATA] = ELFDATA2LSB;
+   ehdr->e_ident[EI_VERSION] = EV_CURRENT;
+
+   ehdr->e_type = ET_CORE;
+   ehdr->e_machine = EM_386;
+   ehdr->e_version = EV_CURRENT;
+   ehdr->e_entry = 0;
+   ehdr->e_phoff = sizeof(Elf32_Ehdr);
+   ehdr->e_shoff = 0;
+   ehdr->e_flags = 0;
+   ehdr->e_ehsize = sizeof(Elf32_Ehdr);
+   ehdr->e_phentsize = sizeof(Elf32_Phdr);
+   ehdr->e_phnum = num_phdrs;
+   ehdr->e_shentsize = 0;
+   ehdr->e_shnum = 0;
+   ehdr->e_shstrndx = 0;
+
+}
+
+static void fill_phdr(Elf32_Phdr *phdr, const Segment *seg, UInt off, Bool write)
+{
+   write = write && should_dump(seg);
+
+   VG_(memset)(phdr, 0, sizeof(*phdr));
+
+   phdr->p_type = PT_LOAD;
+   phdr->p_offset = off;
+   phdr->p_vaddr = seg->addr;
+   phdr->p_paddr = 0;
+   phdr->p_filesz = write ? seg->len : 0;
+   phdr->p_memsz = seg->len;
+   phdr->p_flags = 0;
+
+   if (seg->prot & VKI_PROT_READ)
+      phdr->p_flags |= PF_R;
+   if (seg->prot & VKI_PROT_WRITE)
+      phdr->p_flags |= PF_W;
+   if (seg->prot & VKI_PROT_EXEC)
+      phdr->p_flags |= PF_X;
+
+   phdr->p_align = VKI_BYTES_PER_PAGE;
+}
+
+struct note {
+   struct note *next;
+   Elf32_Nhdr note;
+   Char name[0];
+};
+
+static UInt note_size(const struct note *n)
+{
+   return sizeof(Elf32_Nhdr) + ROUNDUP(VG_(strlen)(n->name)+1, 4) + ROUNDUP(n->note.n_descsz, 4);
+}
+
+static void add_note(struct note **list, const Char *name, UInt type, const void *data, UInt datasz)
+{
+   Int namelen = VG_(strlen)(name)+1;
+   Int notelen = sizeof(struct note) + 
+      ROUNDUP(namelen, 4) + 
+      ROUNDUP(datasz, 4);
+   struct note *n = VG_(arena_malloc)(VG_AR_CORE, notelen);
+
+   VG_(memset)(n, 0, notelen);
+
+   n->next = *list;
+   *list = n;
+
+   n->note.n_type = type;
+   n->note.n_namesz = namelen;
+   n->note.n_descsz = datasz;
+
+   VG_(memcpy)(n->name, name, namelen);
+   VG_(memcpy)(n->name+ROUNDUP(namelen,4), data, datasz);
+}
+
+static void write_note(Int fd, const struct note *n)
+{
+   VG_(write)(fd, &n->note, note_size(n));
+}
+
+static void fill_prpsinfo(const ThreadState *tst, struct elf_prpsinfo *prpsinfo)
+{
+   Char *name;
+
+   VG_(memset)(prpsinfo, 0, sizeof(*prpsinfo));
+
+   switch(tst->status) {
+   case VgTs_Runnable:
+      prpsinfo->pr_sname = 'R';
+      break;
+
+   case VgTs_WaitJoinee:
+      prpsinfo->pr_sname = 'Z';
+      prpsinfo->pr_zomb = 1;
+      break;
+
+   case VgTs_WaitJoiner:
+   case VgTs_WaitMX:
+   case VgTs_WaitCV:
+   case VgTs_WaitSys:
+   case VgTs_Sleeping:
+      prpsinfo->pr_sname = 'S';
+      break;
+
+   case VgTs_Empty:
+      /* ? */
+      break;
+   }
+
+   prpsinfo->pr_uid = 0;
+   prpsinfo->pr_gid = 0;
+   
+   name = VG_(resolve_filename)(VG_(clexecfd));
+
+   if (name != NULL) {
+      Char *n = name+VG_(strlen)(name)-1;
+
+      while(n > name && *n != '/')
+	 n--;
+      if (n != name)
+	 n++;
+
+      VG_(strncpy)(prpsinfo->pr_fname, n, sizeof(prpsinfo->pr_fname));
+   }
+}
+
+static void fill_prstatus(const ThreadState *tst, struct elf_prstatus *prs, const vki_ksiginfo_t *si)
+{
+   struct user_regs_struct *regs;
+
+   VG_(memset)(prs, 0, sizeof(*prs));
+
+   prs->pr_info.si_signo = si->si_signo;
+   prs->pr_info.si_code = si->si_code;
+   prs->pr_info.si_errno = 0;
+
+   prs->pr_cursig = si->si_signo;
+
+   prs->pr_pid = VG_(main_pid) + tst->tid;	/* just to distinguish threads from each other */
+   prs->pr_ppid = 0;
+   prs->pr_pgrp = VG_(main_pgrp);
+   prs->pr_sid = VG_(main_pgrp);
+   
+   regs = (struct user_regs_struct *)prs->pr_reg;
+
+   vg_assert(sizeof(*regs) == sizeof(prs->pr_reg));
+
+   if (VG_(is_running_thread)(tst->tid)) {
+      regs->eflags = VG_(baseBlock)[VGOFF_(m_eflags)];
+      regs->esp = VG_(baseBlock)[VGOFF_(m_esp)];
+      regs->eip = VG_(baseBlock)[VGOFF_(m_eip)];
+
+      regs->ebx = VG_(baseBlock)[VGOFF_(m_ebx)];
+      regs->ecx = VG_(baseBlock)[VGOFF_(m_ecx)];
+      regs->edx = VG_(baseBlock)[VGOFF_(m_edx)];
+      regs->esi = VG_(baseBlock)[VGOFF_(m_esi)];
+      regs->edi = VG_(baseBlock)[VGOFF_(m_edi)];
+      regs->ebp = VG_(baseBlock)[VGOFF_(m_ebp)];
+      regs->eax = VG_(baseBlock)[VGOFF_(m_eax)];
+
+      regs->cs = VG_(baseBlock)[VGOFF_(m_cs)];
+      regs->ds = VG_(baseBlock)[VGOFF_(m_ds)];
+      regs->ss = VG_(baseBlock)[VGOFF_(m_ss)];
+      regs->es = VG_(baseBlock)[VGOFF_(m_es)];
+      regs->fs = VG_(baseBlock)[VGOFF_(m_fs)];
+      regs->gs = VG_(baseBlock)[VGOFF_(m_gs)];
+   } else {
+      regs->eflags = tst->m_eflags;
+      regs->esp = tst->m_esp;
+      regs->eip = tst->m_eip;
+
+      regs->ebx = tst->m_ebx;
+      regs->ecx = tst->m_ecx;
+      regs->edx = tst->m_edx;
+      regs->esi = tst->m_esi;
+      regs->edi = tst->m_edi;
+      regs->ebp = tst->m_ebp;
+      regs->eax = tst->m_eax;
+
+      regs->cs = tst->m_cs;
+      regs->ds = tst->m_ds;
+      regs->ss = tst->m_ss;
+      regs->es = tst->m_es;
+      regs->fs = tst->m_fs;
+      regs->gs = tst->m_gs;
+   }
+}
+
+static void fill_fpu(const ThreadState *tst, elf_fpregset_t *fpu)
+{
+   const Char *from;
+
+   if (VG_(is_running_thread)(tst->tid)) {
+      from = (const Char *)&VG_(baseBlock)[VGOFF_(m_ssestate)];
+   } else {
+      from = (const Char *)&tst->m_sse;
+   }
+
+   if (VG_(have_ssestate)) {
+      UShort *to;
+      Int i;
+
+      /* This is what the kernel does */
+      VG_(memcpy)(fpu, from, 7*sizeof(long));
+   
+      to = (UShort *)&fpu->st_space[0];
+      from += 18 * sizeof(UShort);
+
+      for(i = 0; i < 8; i++, to += 5, from += 8) 
+	 VG_(memcpy)(to, from, 5*sizeof(UShort));
+   } else
+      VG_(memcpy)(fpu, from, sizeof(*fpu));
+}
+
+static void fill_xfpu(const ThreadState *tst, elf_fpxregset_t *xfpu)
+{
+   UShort *from;
+
+   if (VG_(is_running_thread)(tst->tid)) 
+      from = (UShort *)&VG_(baseBlock)[VGOFF_(m_ssestate)];
+   else 
+      from = (UShort *)tst->m_sse;
+
+   VG_(memcpy)(xfpu, from, sizeof(*xfpu));
+}
+
+static void make_coredump(ThreadId tid, const vki_ksiginfo_t *si, UInt max_size)
+{
+   Char buf[1000];
+   Char *basename = "vgcore";
+   Char *coreext = "";
+   Int seq = 0;
+   Int core_fd;
+   Segment *seg;
+   Elf32_Ehdr ehdr;
+   Elf32_Phdr *phdrs;
+   Int num_phdrs;
+   Int i;
+   UInt off;
+   struct note *notelist, *note;
+   UInt notesz;
+   struct elf_prpsinfo prpsinfo;
+   struct elf_prstatus prstatus;
+
+   if (VG_(clo_log_name) != NULL) {
+      coreext = ".core";
+      basename = VG_(clo_log_name);
+   }
+
+   for(;;) {
+      if (seq == 0)
+	 VG_(sprintf)(buf, "%s%s.pid%d",
+		      basename, coreext, VG_(main_pid));
+      else
+	 VG_(sprintf)(buf, "%s%s.pid%d.%d",
+		      basename, coreext, VG_(main_pid), seq);
+      seq++;
+
+      core_fd = VG_(open)(buf, 			   
+			  VKI_O_CREAT|VKI_O_WRONLY|VKI_O_EXCL|VKI_O_TRUNC, 
+			  VKI_S_IRUSR|VKI_S_IWUSR);
+      if (core_fd >= 0)
+	 break;
+
+      if (core_fd != -VKI_EEXIST)
+	 return;		/* can't create file */
+   }
+
+   /* First, count how many memory segments to dump */
+   num_phdrs = 1;		/* start with notes */
+   for(seg = VG_(first_segment)();
+       seg != NULL;
+       seg = VG_(next_segment)(seg)) {
+      if (!may_dump(seg))
+	 continue;
+
+      num_phdrs++;
+   }
+
+   fill_ehdr(&ehdr, num_phdrs);
+
+   /* Second, work out their layout */
+   phdrs = VG_(arena_malloc)(VG_AR_CORE, sizeof(*phdrs) * num_phdrs);
+
+   for(i = 1; i < VG_N_THREADS; i++) {
+      elf_fpregset_t fpu;
+
+      if (VG_(threads)[i].status == VgTs_Empty)
+	 continue;
+
+      if (VG_(have_ssestate)) {
+	 elf_fpxregset_t xfpu;
+
+	 fill_xfpu(&VG_(threads)[i], &xfpu);
+	 add_note(&notelist, "LINUX", NT_PRXFPREG, &xfpu, sizeof(xfpu));
+      }
+
+      fill_fpu(&VG_(threads)[i], &fpu);
+      add_note(&notelist, "CORE", NT_FPREGSET, &fpu, sizeof(fpu));
+
+      fill_prstatus(&VG_(threads)[i], &prstatus, si);
+      add_note(&notelist, "CORE", NT_PRSTATUS, &prstatus, sizeof(prstatus));
+   }
+
+   fill_prpsinfo(&VG_(threads)[tid], &prpsinfo);
+   add_note(&notelist, "CORE", NT_PRPSINFO, &prpsinfo, sizeof(prpsinfo));
+
+   for(note = notelist, notesz = 0; note != NULL; note = note->next)
+      notesz += note_size(note);
+
+   off = sizeof(ehdr) + sizeof(*phdrs) * num_phdrs;
+
+   phdrs[0].p_type = PT_NOTE;
+   phdrs[0].p_offset = off;
+   phdrs[0].p_vaddr = 0;
+   phdrs[0].p_paddr = 0;
+   phdrs[0].p_filesz = notesz;
+   phdrs[0].p_memsz = 0;
+   phdrs[0].p_flags = 0;
+   phdrs[0].p_align = 0;
+
+   off += notesz;
+
+   off = PGROUNDUP(off);
+
+   for(seg = VG_(first_segment)(), i = 1;
+       seg != NULL;
+       seg = VG_(next_segment)(seg), i++) {
+      if (!may_dump(seg))
+	 continue;
+
+      fill_phdr(&phdrs[i], seg, off, (seg->len + off) < max_size);
+      
+      off += phdrs[i].p_filesz;
+   }
+
+   /* write everything out */
+   VG_(write)(core_fd, &ehdr, sizeof(ehdr));
+   VG_(write)(core_fd, phdrs, sizeof(*phdrs) * num_phdrs);
+
+   for(note = notelist; note != NULL; note = note->next)
+      write_note(core_fd, note);
+   
+   VG_(lseek)(core_fd, phdrs[1].p_offset, VKI_SEEK_SET);
+
+   for(seg = VG_(first_segment)(), i = 1;
+       seg != NULL;
+       seg = VG_(next_segment)(seg), i++) {
+      if (!should_dump(seg))
+	 continue;
+
+      vg_assert(VG_(lseek)(core_fd, 0, VKI_SEEK_CUR) == phdrs[i].p_offset);
+      if (phdrs[i].p_filesz > 0)
+	 VG_(write)(core_fd, (void *)seg->addr, seg->len);
+   }
+
+   VG_(close)(core_fd);
+}
+
 /* 
    Perform the default action of a signal.  Returns if the default
    action isn't fatal.
@@ -1328,7 +1717,20 @@ static void vg_default_action(const vki_ksiginfo_t *info, ThreadId tid)
 		   sigNo, terminate ? "terminate" : "", core ? "+core" : "");
 
    if (terminate) {
-      if (VG_(clo_verbosity) != 0 && (core || VG_(clo_verbosity) > 1)) {
+      struct vki_rlimit corelim;
+      Bool could_core = core;
+
+      if (core) {
+	 /* If they set the core-size limit to zero, don't generate a
+	    core file */
+	 
+	 VG_(getrlimit)(VKI_RLIMIT_CORE, &corelim);
+
+	 if (corelim.rlim_cur == 0)
+	    core = False;
+      }
+
+      if (VG_(clo_verbosity) != 0 && (could_core || VG_(clo_verbosity) > 1)) {
 	 VG_(message)(Vg_UserMsg, "");
 	 VG_(message)(Vg_UserMsg, "Process terminating with default action of signal %d (%s)%s", 
 		      sigNo, signame(sigNo), core ? ": dumping core" : "");
@@ -1395,10 +1797,16 @@ static void vg_default_action(const vki_ksiginfo_t *info, ThreadId tid)
          VG_(start_debugger)( tid );
       }
 
-      if (VG_(fatal_signal_set)) {
-	 VG_(fatal_sigNo) = sigNo;
-	 __builtin_longjmp(VG_(fatal_signal_jmpbuf), 1);
+      if (core) {
+	 static struct vki_rlimit zero = { 0, 0 };
+
+	 make_coredump(tid, info, corelim.rlim_cur);
+
+	 /* make sure we don't get a confusing kernel-generated coredump */
+	 VG_(setrlimit)(VKI_RLIMIT_CORE, &zero);
       }
+
+      VG_(scheduler_handle_fatal_signal)( sigNo );
    }
 
    VG_(kill_self)(sigNo);
@@ -1406,6 +1814,38 @@ static void vg_default_action(const vki_ksiginfo_t *info, ThreadId tid)
    vg_assert(!terminate);
 }
 
+static void synth_fault_common(ThreadId tid, Addr addr, Int si_code)
+{
+   vki_ksiginfo_t info;
+
+   vg_assert(VG_(threads)[tid].status == VgTs_Runnable);
+
+   info.si_signo = VKI_SIGSEGV;
+   info.si_code = si_code;
+   info._sifields._sigfault._addr = (void*)addr;
+
+   VG_(resume_scheduler)(VKI_SIGSEGV, &info);
+   VG_(deliver_signal)(tid, &info, False);
+}
+
+// Synthesize a fault where the address is OK, but the page
+// permissions are bad.
+void VG_(synth_fault_perms)(ThreadId tid, Addr addr)
+{
+   synth_fault_common(tid, addr, 2);
+}
+
+// Synthesize a fault where the address there's nothing mapped at the address.
+void VG_(synth_fault_mapping)(ThreadId tid, Addr addr)
+{
+   synth_fault_common(tid, addr, 1);
+}
+
+// Synthesize a misc memory fault.
+void VG_(synth_fault)(ThreadId tid)
+{
+   synth_fault_common(tid, 0, 0x80);
+}
 
 void VG_(deliver_signal) ( ThreadId tid, const vki_ksiginfo_t *info, Bool async )
 {
@@ -1482,7 +1922,7 @@ void VG_(deliver_signal) ( ThreadId tid, const vki_ksiginfo_t *info, Bool async
 	 /* Do the ONESHOT thing. */
 	 handler->scss_handler = VKI_SIG_DFL;
 
-	 VG_(handle_SCSS_change)( False /* lazy update */ );
+	 handle_SCSS_change( False /* lazy update */ );
       }
    
       switch(tst->status) {
@@ -1610,9 +2050,9 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
    /* Sanity check.  Ensure we're really running on the signal stack
       we asked for. */
    if (!(
-         ((Char*)(&(VG_(sigstack)[0])) <= (Char*)(&dummy_local))
+         ((Char*)(&(sigstack[0])) <= (Char*)(&dummy_local))
          &&
-         ((Char*)(&dummy_local) < (Char*)(&(VG_(sigstack)[VG_SIGSTACK_SIZE_W])))
+         ((Char*)(&dummy_local) < (Char*)(&(sigstack[VG_SIGSTACK_SIZE_W])))
          )
       ) {
      VG_(message)(Vg_DebugMsg, 
@@ -1626,8 +2066,8 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
          "rebuild your prog without -p/-pg");
    }
 
-   vg_assert((Char*)(&(VG_(sigstack)[0])) <= (Char*)(&dummy_local));
-   vg_assert((Char*)(&dummy_local) < (Char*)(&(VG_(sigstack)[VG_SIGSTACK_SIZE_W])));
+   vg_assert((Char*)(&(sigstack[0])) <= (Char*)(&dummy_local));
+   vg_assert((Char*)(&dummy_local) < (Char*)(&(sigstack[VG_SIGSTACK_SIZE_W])));
 
    /* Special fault-handling case. We can now get signals which can
       act upon and immediately restart the faulting instruction.
@@ -1642,6 +2082,8 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
       seg = VG_(find_segment)(fault);
       if (seg != NULL)
 	 seg = VG_(next_segment)(seg);
+      else 
+         seg = VG_(first_segment)();
 
       if (VG_(clo_trace_signals)) {
 	 if (seg == NULL)
@@ -1667,15 +2109,13 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
 	    then extend the stack segment. 
 	 */
 	 Addr base = PGROUNDDN(esp);
-	 Char *ret = VG_(mmap)((Char *)base, seg->addr - base, 
-			       VKI_PROT_READ | VKI_PROT_WRITE | VKI_PROT_EXEC,
-			       VKI_MAP_PRIVATE | VKI_MAP_FIXED | VKI_MAP_ANONYMOUS | VKI_MAP_CLIENT,
-			       -1, 0);
-	 if ((Addr)ret == base) {
-	    VG_(map_segment)(base, seg->addr - base,
-			     VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC,
-			     SF_STACK|SF_GROWDOWN);
-	    return;		/* restart instruction */
+         if ((void*)-1 != VG_(mmap)((Char *)base, seg->addr - base,
+                              VKI_PROT_READ|VKI_PROT_WRITE|VKI_PROT_EXEC,
+                              VKI_MAP_PRIVATE|VKI_MAP_FIXED|VKI_MAP_ANONYMOUS|VKI_MAP_CLIENT,
+                              SF_STACK|SF_GROWDOWN,
+                              -1, 0))
+         {
+           return;             // extension succeeded, restart instruction
 	 }
 	 /* Otherwise fall into normal signal handling */
       } else if (info->si_code == 2 && /* SEGV_ACCERR */
@@ -1700,14 +2140,9 @@ void vg_sync_signalhandler ( Int sigNo, vki_ksiginfo_t *info, struct vki_ucontex
       }
    }
 
-   if (VG_(scheduler_jmpbuf_valid)) {
-      /* Can't continue; must longjmp back to the scheduler and thus
-         enter the sighandler immediately. */
-      VG_(memcpy)(&VG_(unresumable_siginfo), info, sizeof(vki_ksiginfo_t));
-   
-      VG_(longjmpd_on_signal) = sigNo;
-      __builtin_longjmp(VG_(scheduler_jmpbuf),1);
-   }
+   /* Can't continue; must longjmp back to the scheduler and thus
+      enter the sighandler immediately. */
+   VG_(resume_scheduler)(sigNo, info);
 
    if (info->si_code <= VKI_SI_USER) {
       /* 
@@ -1926,17 +2361,13 @@ void VG_(sigstartup_actions) ( void )
    vg_scss.scss_per_sig[VKI_SIGVGKILL].scss_flags = VKI_SA_SIGINFO;
    VG_(ksigfillset)(&vg_scss.scss_per_sig[VKI_SIGVGKILL].scss_mask);
 
-   /* Copy the alt stack, if any. */
-   ret = VG_(ksigaltstack)(NULL, &vg_scss.altstack);
-   vg_assert(ret == 0);
-
    /* Copy the process' signal mask into the root thread. */
    vg_assert(VG_(threads)[1].status == VgTs_Runnable);
    VG_(threads)[1].sig_mask = saved_procmask;
    VG_(proxy_setsigmask)(1);
 
    /* Register an alternative stack for our own signal handler to run on. */
-   altstack_info.ss_sp = &(VG_(sigstack)[0]);
+   altstack_info.ss_sp = &(sigstack[0]);
    altstack_info.ss_size = VG_SIGSTACK_SIZE_W * sizeof(UInt);
    altstack_info.ss_flags = 0;
    ret = VG_(ksigaltstack)(&altstack_info, NULL);
@@ -1954,56 +2385,10 @@ void VG_(sigstartup_actions) ( void )
 
    /* Calculate SKSS and apply it.  This also sets the initial kernel
       mask we need to run with. */
-   VG_(handle_SCSS_change)( True /* forced update */ );
+   handle_SCSS_change( True /* forced update */ );
 
 }
 
-
-/* Copy the process' sim signal state to the real state,
-   for when we transfer from the simulated to real CPU.
-   PROBLEM: what if we're running a signal handler when we
-   get here?  Hmm.
-   I guess we wind up in vg_signalreturn_bogusRA, *or* the
-   handler has done/will do a longjmp, in which case we're ok.
-*/
-void VG_(sigshutdown_actions) ( void )
-{
-   Int i, ret;
-
-   vki_ksigset_t  saved_procmask;
-   vki_ksigaction sa;
-
-   VG_(block_all_host_signals)( &saved_procmask );
-
-   /* Copy per-signal settings from SCSS. */
-   for (i = 1; i <= VKI_KNSIG; i++) {
-
-      sa.ksa_handler  = vg_scss.scss_per_sig[i].scss_handler;
-      sa.ksa_flags    = vg_scss.scss_per_sig[i].scss_flags;
-      sa.ksa_mask     = vg_scss.scss_per_sig[i].scss_mask;
-      sa.ksa_restorer = vg_scss.scss_per_sig[i].scss_restorer;
-
-      if (VG_(clo_trace_signals))
-         VG_(printf)("restoring handler 0x%x for signal %d\n", 
-                     (Addr)(sa.ksa_handler), i );
-
-      /* Set the old host action */
-      ret = VG_(ksigaction)(i, &sa, NULL);
-      if (i != VKI_SIGKILL && i != VKI_SIGSTOP) 
-         vg_assert(ret == 0);
-   }
-
-   /* Restore the sig alt stack. */
-   ret = VG_(ksigaltstack)(&vg_scss.altstack, NULL);
-   vg_assert(ret == 0);
-
-   /* A bit of a kludge -- set the sigmask to that of the root
-      thread. */
-   vg_assert(VG_(threads)[1].status != VgTs_Empty);
-   VG_(restore_all_host_signals)( &VG_(threads)[1].sig_mask );
-}
-
-
 /*--------------------------------------------------------------------*/
 /*--- end                                             vg_signals.c ---*/
 /*--------------------------------------------------------------------*/
