@@ -34,6 +34,8 @@
 
 #include <elf.h>          /* ELF defns                      */
 
+static Bool
+intercept_demangle(const Char*, Char*, Int);
 
 /* Majorly rewritten Sun 3 Feb 02 to enable loading symbols from
    dlopen()ed libraries, which is something that KDE3 does a lot.
@@ -213,7 +215,7 @@ void VG_(addLineInfo) ( SegInfo* si,
     * correspond to one line, unlike with C-level debug info where
     * multiple instructions can map to the one line), but avoid
     * catching any other instructions bogusly. */
-   if (this > next) {
+   if (this > next && VG_(clo_verbosity) > 2) {
        VG_(message)(Vg_DebugMsg, 
                     "warning: line info addresses out of order "
                     "at entry %d: 0x%x 0x%x", entry, this, next);
@@ -395,7 +397,9 @@ static Int compare_RiSym(void *va, void *vb) {
    RiSym *a = (RiSym *)va;
    RiSym *b = (RiSym *)vb;
    
-   return a->addr - b->addr;
+   if (a->addr < b->addr) return -1;
+   if (a->addr > b->addr) return  1;
+   return 0;
 }
 
 /* Two symbols have the same address.  Which name do we prefer?
@@ -441,6 +445,17 @@ static RiSym *prefersym(RiSym *a, RiSym *b)
       lenb = lt;
    }
 
+   /* Ugh.  If we get a "free", always choose it.  This is because
+      normally, this function would choose "cfree" over free.  cfree is
+      an alias for free.  If there's any more symbols like this, we may
+      want to consider making this mechanism more generic.
+    */
+   if(VG_(strcmp)(a->name, "free") == 0)
+       return a;
+
+   if(VG_(strcmp)(b->name, "free") == 0)
+       return b;
+
    for(i = pfx = 0; i < sizeof(prefixes)/sizeof(*prefixes); i++) {
       Int pfxlen = prefixes[i].len;
 
@@ -471,6 +486,18 @@ void canonicaliseSymtab ( SegInfo* si )
 
    VG_(ssort)(si->symtab, si->symtab_used, sizeof(*si->symtab), compare_RiSym);
 
+   for (i = 0; i < si->symtab_used; i++) {
+      if(VG_(strncmp)(si->symtab[i].name, VG_INTERCEPT_PREFIX,
+                      VG_INTERCEPT_PREFIX_LEN) == 0) {
+         int len = VG_(strlen)(si->symtab[i].name);
+         char *buf = VG_(malloc)(len), *colon;
+         intercept_demangle(si->symtab[i].name, buf, len);
+	 colon = buf + VG_(strlen)(buf) - 1;
+	 while(*colon != ':') colon--;
+	 VG_(strncpy_safely)(si->symtab[i].name, colon+1, len);
+      }
+   }
+
   cleanup_more:
  
    /* If two symbols have identical address ranges, favour the
@@ -579,7 +606,9 @@ static Int compare_ScopeRange(void *va, void *vb) {
    ScopeRange *a = (ScopeRange *)va;
    ScopeRange *b = (ScopeRange *)vb;
    
-   return a->addr - b->addr;
+   if (a->addr < b->addr) return -1;
+   if (a->addr > b->addr) return  1;
+   return 0;
 }
 
 static 
@@ -648,7 +677,9 @@ static Int compare_RiLoc(void *va, void *vb) {
    RiLoc *a = (RiLoc *)va;
    RiLoc *b = (RiLoc *)vb;
 
-   return a->addr - b->addr;
+   if (a->addr < b->addr) return -1;
+   if (a->addr > b->addr) return  1;
+   return 0;
 }
 
 static 
@@ -746,9 +777,89 @@ Bool VG_(is_object_file)(const void *buf)
    return False;
 }
 
+/*
+ * Demangle an intercept symbol into library:func form
+ */
+
+static Bool
+intercept_demangle(const Char* symbol, Char* result, Int nbytes)
+{
+   int i, j = 0;
+   int len = VG_(strlen)(symbol);
+
+   for(i = VG_INTERCEPT_PREFIX_LEN; i < len; i++) {
+      if(symbol[i] == '$') {
+         i++;
+         if(symbol[i] == '$') {
+            result[j] = '$';
+         } else if((symbol[i] >= '0' && symbol[i] <= '9') ||
+                   (symbol[i] >= 'a' && symbol[i] <= 'f') ||
+                   (symbol[i] >= 'A' && symbol[i] <= 'F')) {
+            int x = symbol[i++];
+            int y = symbol[i];
+            if(x >= '0' && x <= '9') {
+               x -= '0';
+            } else if(x >= 'a' && x <= 'f') {
+               x -= 'a';
+            } else if(x >= 'A' && x <= 'F') {
+               x -= 'A';
+            }
+            if(y >= '0' && y <= '9') {
+               y -= '0';
+            } else if(y >= 'a' && y <= 'f') {
+               y = y - 'a' + 10;
+            } else if(y >= 'A' && y <= 'F') {
+               y = y - 'A' + 10;
+            } else {
+               return False;
+            }
+            result[j] = (x << 4) | y;
+         } else {
+            return False;
+         }
+      } else {
+         result[j] = symbol[i];
+      }
+      if(j >= nbytes) {
+         result[j] = '\0';
+         return True;
+      }
+      j++;
+   }
+   result[j] = '\0';
+   return True;
+}
+
+// Forward declaration
+static void add_redirect_addr(const Char *from_lib, const Char *from_sym,
+			      Addr to_addr);
+
+static
+void handle_intercept( SegInfo* si, Char* symbol, Elf32_Sym* sym)
+{
+   int len = VG_(strlen)(symbol) + 1 - VG_INTERCEPT_PREFIX_LEN;
+   char *lib = VG_(malloc)(len);
+   Char *func;
+
+   intercept_demangle(symbol, lib, len);
+   func = lib + VG_(strlen)(lib)-1;
+
+   while(*func != ':') func--;
+   *func = '\0';
+
+   add_redirect_addr(lib, func+1, si->offset + sym->st_value);
+   VG_(free)(lib);
+}
+
+static
+void handle_wrapper( SegInfo* si, Char* symbol, Elf32_Sym* sym)
+{
+   VG_(intercept_libc_freeres_wrapper)((Addr)(si->offset + sym->st_value));
+}
+
 /* Read a symbol table (normal or dynamic) */
 static
-void read_symtab( SegInfo* si, Char* tab_name,
+void read_symtab( SegInfo* si, Char* tab_name, Bool do_intercepts,
                   Elf32_Sym* o_symtab, UInt o_symtab_sz,
                   UChar*     o_strtab, UInt o_strtab_sz )
 {
@@ -805,6 +916,22 @@ void read_symtab( SegInfo* si, Char* tab_name,
       }               
 #     endif
 
+      /*
+       * Is this symbol a magic valgrind-intercept symbol?  If so,
+       * hand this off to the interceptinator.
+       */
+      if (do_intercepts) {
+         if (VG_(strncmp)((Char*)o_strtab+sym->st_name,
+                          VG_INTERCEPT_PREFIX,
+			  VG_INTERCEPT_PREFIX_LEN) == 0) {
+           handle_intercept(si, (Char*)o_strtab+sym->st_name, sym);
+         } else if (VG_(strncmp)((Char*)o_strtab+sym->st_name,
+                                 VG_WRAPPER_PREFIX,
+				 VG_WRAPPER_PREFIX_LEN) == 0) {
+           handle_wrapper(si, (Char*)o_strtab+sym->st_name, sym);
+         }
+      }
+
       /* Figure out if we're interested in the symbol.
          Firstly, is it of the right flavour?  */
       if ( ! ( (ELF32_ST_BIND(sym->st_info) == STB_GLOBAL ||
@@ -886,6 +1013,147 @@ void read_symtab( SegInfo* si, Char* tab_name,
    }
 }
 
+/*
+ * This routine for calculating the CRC for a separate debug file
+ * is GPLed code borrowed from binutils.
+ */
+static UInt
+calc_gnu_debuglink_crc32(UInt crc, const UChar *buf, Int len)
+{
+  static const UInt crc32_table[256] =
+    {
+      0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419,
+      0x706af48f, 0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4,
+      0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07,
+      0x90bf1d91, 0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
+      0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7, 0x136c9856,
+      0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
+      0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4,
+      0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
+      0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3,
+      0x45df5c75, 0xdcd60dcf, 0xabd13d59, 0x26d930ac, 0x51de003a,
+      0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599,
+      0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
+      0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190,
+      0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f,
+      0x9fbfe4a5, 0xe8b8d433, 0x7807c9a2, 0x0f00f934, 0x9609a88e,
+      0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
+      0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed,
+      0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
+      0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3,
+      0xfbd44c65, 0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
+      0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a,
+      0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5,
+      0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa, 0xbe0b1010,
+      0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
+      0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17,
+      0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6,
+      0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615,
+      0x73dc1683, 0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
+      0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1, 0xf00f9344,
+      0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
+      0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a,
+      0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
+      0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1,
+      0xa6bc5767, 0x3fb506dd, 0x48b2364b, 0xd80d2bda, 0xaf0a1b4c,
+      0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef,
+      0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
+      0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe,
+      0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31,
+      0x2cd99e8b, 0x5bdeae1d, 0x9b64c2b0, 0xec63f226, 0x756aa39c,
+      0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
+      0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b,
+      0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
+      0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1,
+      0x18b74777, 0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
+      0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45, 0xa00ae278,
+      0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7,
+      0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc, 0x40df0b66,
+      0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
+      0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605,
+      0xcdd70693, 0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8,
+      0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b,
+      0x2d02ef8d
+    };
+  const UChar *end;
+
+  crc = ~crc & 0xffffffff;
+  for (end = buf + len; buf < end; ++ buf)
+    crc = crc32_table[(crc ^ *buf) & 0xff] ^ (crc >> 8);
+  return ~crc & 0xffffffff;;
+}
+
+/*
+ * Try and open a separate debug file, ignoring any where the CRC does
+ * not match the value from the main object file.
+ */
+static
+Addr open_debug_file( Char* name, UInt crc, UInt* size )
+{
+   Int fd;
+   struct vki_stat stat_buf;
+   Addr addr;
+
+   if ((fd = VG_(open)(name, VKI_O_RDONLY, 0)) < 0)
+      return 0;
+
+   if (VG_(fstat)(fd, &stat_buf) != 0) {
+      VG_(close)(fd);
+      return 0;
+   }
+
+   *size = stat_buf.st_size;
+   
+   if ((addr = (Addr)VG_(mmap)(NULL, *size, VKI_PROT_READ,
+                               VKI_MAP_PRIVATE|VKI_MAP_NOSYMS, 
+                               0, fd, 0)) == (Addr)-1) 
+   {
+      VG_(close)(fd);
+      return 0;
+   }
+
+   VG_(close)(fd);
+   
+   if (calc_gnu_debuglink_crc32(0, (UChar*)addr, *size) != crc) {
+      int res = VG_(munmap)((void*)addr, *size);
+      vg_assert(0 == res);
+      return 0;
+   }
+   
+   return addr;
+}
+
+/*
+ * Try to find a separate debug file for a given object file.
+ */
+static
+Addr find_debug_file( Char* objpath, Char* debugname, UInt crc, UInt* size )
+{
+   Char *objdir = VG_(strdup)(objpath);
+   Char *objdirptr;
+   Char *debugpath;
+   Addr addr = 0;
+  
+   if ((objdirptr = VG_(strrchr)(objdir, '/')) != NULL)
+      *objdirptr = '\0';
+
+   debugpath = VG_(malloc)(VG_(strlen)(objdir) + VG_(strlen)(debugname) + 16);
+   
+   VG_(sprintf)(debugpath, "%s/%s", objdir, debugname);
+
+   if ((addr = open_debug_file(debugpath, crc, size)) == 0) {
+      VG_(sprintf)(debugpath, "%s/.debug/%s", objdir, debugname);
+      if ((addr = open_debug_file(debugpath, crc, size)) == 0) {
+         VG_(sprintf)(debugpath, "/usr/lib/debug%s/%s", objdir, debugname);
+         addr = open_debug_file(debugpath, crc, size);
+      }
+   }
+
+   VG_(free)(debugpath);
+   VG_(free)(objdir);
+   
+   return addr;
+}
 
 /* Read the symbols from the object/exe specified by the SegInfo into
    the tables within the supplied SegInfo.  */
@@ -901,6 +1169,8 @@ Bool vg_read_lib_symbols ( SegInfo* si )
    Bool          ok;
    Addr          oimage;
    UInt          n_oimage;
+   Addr          dimage = 0;
+   UInt          n_dimage = 0;
    struct vki_stat stat_buf;
 
    oimage = (Addr)NULL;
@@ -925,13 +1195,16 @@ Bool vg_read_lib_symbols ( SegInfo* si )
    }
 
    oimage = (Addr)VG_(mmap)( NULL, n_oimage, 
-                             VKI_PROT_READ, VKI_MAP_PRIVATE|VKI_MAP_NOSYMS, fd, 0 );
+                             VKI_PROT_READ, VKI_MAP_PRIVATE|VKI_MAP_NOSYMS, 
+                             0, fd, 0 );
 
    VG_(close)(fd);
 
    if (oimage == ((Addr)(-1))) {
       VG_(message)(Vg_UserMsg,
-                   "mmap failed on %s", si->filename );
+                   "warning: mmap failed on %s", si->filename );
+      VG_(message)(Vg_UserMsg,
+                   "         no symbols or debug info loaded" );
       return False;
    }
 
@@ -1082,6 +1355,7 @@ Bool vg_read_lib_symbols ( SegInfo* si )
       Elf32_Sym* o_symtab     = NULL; /* .symtab */
       UChar*     o_dynstr     = NULL; /* .dynstr */
       Elf32_Sym* o_dynsym     = NULL; /* .dynsym */
+      Char*      debuglink    = NULL; /* .gnu_debuglink */
       UChar*     stab         = NULL; /* .stab         (stabs)  */
       UChar*     stabstr      = NULL; /* .stabstr      (stabs)  */
       UChar*     debug_line   = NULL; /* .debug_line   (dwarf2) */
@@ -1093,6 +1367,7 @@ Bool vg_read_lib_symbols ( SegInfo* si )
       UInt       o_symtab_sz     = 0;
       UInt       o_dynstr_sz     = 0;
       UInt       o_dynsym_sz     = 0;
+      UInt       debuglink_sz    = 0;
       UInt       stab_sz         = 0;
       UInt       stabstr_sz      = 0;
       UInt       debug_line_sz   = 0;
@@ -1127,6 +1402,8 @@ Bool vg_read_lib_symbols ( SegInfo* si )
          else FIND(".symtab",       o_symtab,     o_symtab_sz,   0, Elf32_Sym*)
          else FIND(".strtab",       o_strtab,     o_strtab_sz,   0, UChar*)
 
+         else FIND(".gnu_debuglink", debuglink,   debuglink_sz,  0, Char*)
+
          else FIND(".stab",         stab,         stab_sz,       0, UChar*)
          else FIND(".stabstr",      stabstr,      stabstr_sz,    0, UChar*)
          else FIND(".debug_line",   debug_line,   debug_line_sz, 0, UChar*)
@@ -1143,14 +1420,67 @@ Bool vg_read_lib_symbols ( SegInfo* si )
          vg_assert((o_symtab_sz % sizeof(Elf32_Sym)) == 0);
       }
 
-      read_symtab(si, "symbol table", 
+      read_symtab(si, "symbol table", False,
                   o_symtab, o_symtab_sz,
                   o_strtab, o_strtab_sz);
 
-      read_symtab(si, "dynamic symbol table",
+      read_symtab(si, "dynamic symbol table", True,
                   o_dynsym, o_dynsym_sz,
                   o_dynstr, o_dynstr_sz);
 
+      /* Did we find a debuglink section? */
+      if (debuglink != NULL) {
+         UInt crc_offset = (VG_(strlen)(debuglink) + 4) & ~3;
+         UInt crc;
+
+         vg_assert(crc_offset + sizeof(UInt) <= debuglink_sz);
+
+         /* Extract the CRC from the debuglink section */
+         crc = *(UInt *)(debuglink + crc_offset);
+
+         /* See if we can find a matching debug file */
+         if ((dimage = find_debug_file(si->filename, debuglink, crc, &n_dimage)) != 0) {
+            ehdr = (Elf32_Ehdr*)dimage;
+
+            if (n_dimage >= sizeof(Elf32_Ehdr) && VG_(is_object_file)(ehdr))
+            {
+               shdr = (Elf32_Shdr*)(dimage + ehdr->e_shoff);
+               sh_strtab = (UChar*)(dimage + shdr[ehdr->e_shstrndx].sh_offset);
+
+               /* Find all interesting sections */
+               for (i = 0; i < ehdr->e_shnum; i++) {
+#                 define FIND(sec_name, sec_data, sec_size, type) \
+                  if (0 == VG_(strcmp)(sec_name, sh_strtab + shdr[i].sh_name)) { \
+                     if (0 != sec_data) \
+                        VG_(core_panic)("repeated section!\n"); \
+                     sec_data = (type)(dimage + shdr[i].sh_offset); \
+                     sec_size = shdr[i].sh_size; \
+                     TRACE_SYMTAB( "%18s: %p .. %p\n", \
+                                   sec_name, sec_data, sec_data + sec_size - 1); \
+                     if ( shdr[i].sh_offset + sec_size > n_dimage ) { \
+                        VG_(symerr)("   section beyond image end?!"); \
+                        goto out; \
+                     } \
+                  }
+
+                  /* Nb: must find where .got and .plt sections will be in the
+                   * executable image, not in the object image transiently loaded. */
+                       FIND(".stab",         stab,         stab_sz,       UChar*)
+                  else FIND(".stabstr",      stabstr,      stabstr_sz,    UChar*)
+                  else FIND(".debug_line",   debug_line,   debug_line_sz, UChar*)
+                  else FIND(".debug",        dwarf1d,      dwarf1d_sz,    UChar*)
+                  else FIND(".line",         dwarf1l,      dwarf1l_sz,    UChar*)
+
+#                 undef FIND
+         
+                  /* Check some sizes */
+                  vg_assert((o_dynsym_sz % sizeof(Elf32_Sym)) == 0);
+                  vg_assert((o_symtab_sz % sizeof(Elf32_Sym)) == 0);
+               }
+            }
+         }
+      }
+
       /* Read the stabs and/or dwarf2 debug information, if any. */
       if (stab != NULL && stabstr != NULL) {
          has_debuginfo = True;
@@ -1173,10 +1503,17 @@ Bool vg_read_lib_symbols ( SegInfo* si )
    }
    res = True;
 
-  out:
-   /* Last, but not least, heave the oimage back overboard. */
-   VG_(munmap) ( (void*)oimage, n_oimage );
+  out: {
+   Int m_res;
+   /* Last, but not least, heave the image(s) back overboard. */
+   if (dimage) {
+      m_res = VG_(munmap) ( (void*)dimage, n_dimage );
+      vg_assert(0 == m_res);
+   }
+   m_res = VG_(munmap) ( (void*)oimage, n_oimage );
+   vg_assert(0 == m_res);
    return res;
+  } 
 }
 
 /*------------------------------------------------------------*/
@@ -1262,7 +1599,7 @@ SegInfo *VG_(read_seg_symbols) ( Segment *seg )
    accuracy of error messages, but we need to do it in order to
    maintain the no-overlapping invariant.
 */
-void VG_(unload_symbols) ( Addr start, UInt length )
+static void unload_symbols ( Addr start, UInt length )
 {
    SegInfo *prev, *curr;
 
@@ -1300,7 +1637,7 @@ void VG_(symtab_decref)(SegInfo *si, Addr start, UInt len)
 {
    vg_assert(si->ref >= 1);
    if (--si->ref == 0)
-      VG_(unload_symbols)(si->start, si->size);
+      unload_symbols(si->start, si->size);
 }
 
 void VG_(symtab_incref)(SegInfo *si)
@@ -2116,8 +2453,8 @@ static Bool resolve_redir_allsegs(CodeRedirect *redir)
 }
 
 /* Redirect a lib/symbol reference to a function at lib/symbol */
-void VG_(add_redirect_sym)(const Char *from_lib, const Char *from_sym,
-			   const Char *to_lib, const Char *to_sym)
+static void add_redirect_sym(const Char *from_lib, const Char *from_sym,
+			     const Char *to_lib, const Char *to_sym)
 {
    CodeRedirect *redir = VG_(SkipNode_Alloc)(&sk_resolved_redir);
 
@@ -2142,8 +2479,8 @@ void VG_(add_redirect_sym)(const Char *from_lib, const Char *from_sym,
 }
 
 /* Redirect a lib/symbol reference to a function at lib/symbol */
-void VG_(add_redirect_addr)(const Char *from_lib, const Char *from_sym,
-			    Addr to_addr)
+static void add_redirect_addr(const Char *from_lib, const Char *from_sym,
+			      Addr to_addr)
 {
    CodeRedirect *redir = VG_(SkipNode_Alloc)(&sk_resolved_redir);
 
@@ -2189,23 +2526,29 @@ void VG_(setup_code_redirect_table) ( void )
    Int i;
 
    for(i = 0; i < sizeof(redirects)/sizeof(*redirects); i++) {
-      VG_(add_redirect_sym)("soname:libc.so.6",		redirects[i].from,
-			    "soname:libpthread.so.0",	redirects[i].to);
+      add_redirect_sym("soname:libc.so.6",	 redirects[i].from,
+                       "soname:libpthread.so.0", redirects[i].to);
    }
 
+   /* Redirect _dl_sysinfo_int80, which is glibc's default system call
+      routine, to the routine in our trampoline page so that the
+      special sysinfo unwind hack in vg_execontext.c will kick in.
+   */
+   add_redirect_addr("soname:ld-linux.so.2", "_dl_sysinfo_int80",
+                     VG_(client_trampoline_code)+VG_(tramp_syscall_offset));
+   
    /* Overenthusiastic use of PLT bypassing by the glibc people also
       means we need to patch the following functions to our own
       implementations of said, in mac_replace_strmem.c.
     */
-   VG_(add_redirect_sym)("soname:libc.so.6", "stpcpy",
-			 "*vgpreload_memcheck.so*", "stpcpy");
-   VG_(add_redirect_sym)("soname:libc.so.6", "strnlen",
-			 "*vgpreload_memcheck.so*", "strnlen");
-
-   VG_(add_redirect_sym)("soname:ld-linux.so.2", "stpcpy",
-			 "*vgpreload_memcheck.so*", "stpcpy");
-   VG_(add_redirect_sym)("soname:ld-linux.so.2", "strchr",
-			 "*vgpreload_memcheck.so*", "strchr");
+   add_redirect_sym("soname:libc.so.6", "stpcpy",
+                    "*vgpreload_memcheck.so*", "stpcpy");
+   add_redirect_sym("soname:libc.so.6", "strnlen",
+                    "*vgpreload_memcheck.so*", "strnlen");
+   add_redirect_sym("soname:ld-linux.so.2", "stpcpy",
+                    "*vgpreload_memcheck.so*", "stpcpy");
+   add_redirect_sym("soname:ld-linux.so.2", "strchr",
+                    "*vgpreload_memcheck.so*", "strchr");
 }
 
 /*------------------------------------------------------------*/
