@@ -37,28 +37,53 @@
 
 #define dis       VG_(print_codegen)
 
+/*------------------------------------------------------------*/
+/*--- Reg-alloc stats                                      ---*/
+/*------------------------------------------------------------*/
+
+static UInt n_uinstrs_prealloc;           // # uinstrs input to reg-alloc
+static UInt n_uinstrs_spill;              // # uinstrs added due to spill code
+static UInt n_translations_needing_spill; // # bbs requiring spill code
+static UInt n_total_reg_rank; // total of register ranks over all translations
+
+void VG_(print_reg_alloc_stats)(void)
+{
+   VG_(message)(Vg_DebugMsg, 
+                "reg-alloc: %d t-req-spill, %d+%d orig+spill uis,",
+                n_translations_needing_spill, 
+                n_uinstrs_prealloc, n_uinstrs_spill );
+   VG_(message)(Vg_DebugMsg, 
+                "           %d total-reg-rank",
+                n_total_reg_rank );
+}
 
 /*------------------------------------------------------------*/
 /*--- Basics                                               ---*/
 /*------------------------------------------------------------*/
 
-/* This one is called by the core */
-UCodeBlock* VG_(alloc_UCodeBlock) ( void )
+#define VG_IS_FLAG_SUBSET(set1,set2) \
+   (( ((FlagSet)set1) & ((FlagSet)set2) ) == ((FlagSet)set1) )
+
+#define VG_UNION_FLAG_SETS(set1,set2) \
+   ( ((FlagSet)set1) | ((FlagSet)set2) )
+
+// This one is local.
+static UCodeBlock* alloc_UCodeBlock ( Addr orig_eip )
 {
    UCodeBlock* cb = VG_(arena_malloc)(VG_AR_CORE, sizeof(UCodeBlock));
-   cb->used = cb->size = cb->nextTemp = 0;
-   cb->instrs = NULL;
+   cb->orig_eip = orig_eip;
+   cb->used     = 0;
+   cb->size     = 0;
+   cb->instrs   = NULL;
+   cb->nextTemp = 0;
    return cb;
 }
 
-/* This one is called by skins */
+// This one is called by tools.
 UCodeBlock* VG_(setup_UCodeBlock) ( UCodeBlock* cb_in )
 {
-   UCodeBlock* cb = VG_(arena_malloc)(VG_AR_CORE, sizeof(UCodeBlock));
-   cb->orig_eip = cb_in->orig_eip;
-   cb->used = cb->size = 0;
+   UCodeBlock* cb = alloc_UCodeBlock( cb_in->orig_eip );
    cb->nextTemp = cb_in->nextTemp;
-   cb->instrs = NULL;
    return cb;
 }
 
@@ -199,23 +224,6 @@ void VG_(copy_UInstr) ( UCodeBlock* cb, UInstr* instr )
    cb->used++;
 }
 
-/* Copy auxiliary info from one uinstr to another. */
-static __inline__ 
-void copyAuxInfoFromTo ( UInstr* src, UInstr* dst )
-{
-   dst->cond            = src->cond;
-   dst->extra4b         = src->extra4b;
-   dst->signed_widen    = src->signed_widen;
-   dst->jmpkind         = src->jmpkind;
-   dst->flags_r         = src->flags_r;
-   dst->flags_w         = src->flags_w;
-   dst->argc            = src->argc;
-   dst->regparms_n      = src->regparms_n;
-   dst->has_ret_val     = src->has_ret_val;
-   dst->regs_live_after = src->regs_live_after;
-}
-
-
 /* Set the lit32 field of the most recent uinsn. */
 void VG_(set_lit_field) ( UCodeBlock* cb, UInt lit32 )
 {
@@ -251,6 +259,17 @@ void VG_(set_flag_fields) ( UCodeBlock* cb,
    LAST_UINSTR(cb).flags_w = uw;
 }
 
+void VG_(set_cond_field) ( UCodeBlock* cb, Condcode cond )
+{
+   LAST_UINSTR(cb).cond = cond;
+}
+
+void VG_(set_widen_fields) ( UCodeBlock* cb, UInt szs, Bool is_signed )
+{
+   LAST_UINSTR(cb).extra4b = szs;
+   LAST_UINSTR(cb).signed_widen = is_signed;
+}
+
 
 Bool VG_(any_flag_use) ( UInstr* u )
 {
@@ -322,11 +341,36 @@ Int VG_(realreg_to_rank) ( Int realReg )
    }
 }
 
+Int VG_(get_num_instrs) ( UCodeBlock* cb )
+{
+   return cb->used;
+}
+
+Int VG_(get_num_temps) ( UCodeBlock* cb )
+{
+   return cb->nextTemp;
+}
+
+UInstr* VG_(get_instr) ( UCodeBlock* cb, Int i )
+{
+   return & cb->instrs[i];
+}
+
+UInstr* VG_(get_last_instr) ( UCodeBlock* cb )
+{
+   return & cb->instrs[cb->used-1];
+}
+   
 
 /*------------------------------------------------------------*/
 /*--- Sanity checking uinstrs.                             ---*/
 /*------------------------------------------------------------*/
 
+// Global variables that indicate where we are in the translation of a basic
+// block, and affect exactly how UInstrs are sanity-checked.
+static Bool beforeRA = True;
+static Bool beforeLiveness  = True;
+
 /* This seems as good a place as any to record some important stuff
    about ucode semantics.
 
@@ -399,7 +443,7 @@ Int VG_(realreg_to_rank) ( Int realReg )
    The size field should be 0 for insns for which it is meaningless,
    ie those which do not directly move/operate on data.
 */
-Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
+static Bool is_sane_UInstr ( UInstr* u )
 {
 #  define LIT0 (u->lit32 == 0)
 #  define LIT8 (((u->lit32) & 0xFFFFFF00) == 0)
@@ -421,7 +465,7 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
 #  define SZf  (  u->size ==  4 || u->size ==  8 || u->size ==   2     \
                || u->size == 10 || u->size == 28 || u->size == 108)
 #  define SZ4m ((u->tag1 == TempReg || u->tag1 == RealReg) \
-                      ? (u->size == 4) : True)
+                      ? (u->size == 4) : SZi)
 
 /* For these ones, two cases:
  *
@@ -474,6 +518,7 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
 
 #  define COND0    (u->cond         == 0)
 #  define EXTRA4b0 (u->extra4b      == 0)
+#  define EXTRA4b12 (u->extra4b     == 1 || u->extra4b == 2)
 #  define SG_WD0   (u->signed_widen == 0)
 #  define JMPKIND0 (u->jmpkind      == 0)
 #  define CCALL0   (u->argc==0 && u->regparms_n==0 && u->has_ret_val==0 && \
@@ -482,8 +527,8 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
                     : True ))
 
 #  define XCONDi   (         EXTRA4b0 && SG_WD0 && JMPKIND0 && CCALL0)
-#  define Xextra4b (COND0             && SG_WD0 && JMPKIND0 && CCALL0)
-#  define XWIDEN   (COND0                       && JMPKIND0 && CCALL0)
+#  define XLEA2    (COND0             && SG_WD0 && JMPKIND0 && CCALL0)
+#  define XWIDEN   (COND0 && EXTRA4b12          && JMPKIND0 && CCALL0)
 #  define XJMP     (                     SG_WD0             && CCALL0)
 #  define XCCALL   (COND0 && EXTRA4b0 && SG_WD0 && JMPKIND0          )
 #  define XOTHER   (COND0 && EXTRA4b0 && SG_WD0 && JMPKIND0 && CCALL0)
@@ -548,21 +593,22 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
    case FPU_W:  return LIT0 && SZf  && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
    case FPU:    return LIT0 && SZ0  && CCf &&  Ls1 &&  N2 &&  N3 && XOTHER;
    case LEA1:   return /*any*/ SZ4  && CC0 &&  TR1 && TR2 &&  N3 && XOTHER;
-   case LEA2:   return /*any*/ SZ4  && CC0 &&  TR1 && TR2 && TR3 && Xextra4b;
+   case LEA2:   return /*any*/ SZ4  && CC0 &&  TR1 && TR2 && TR3 && XLEA2;
    case INCEIP: return LIT0 && SZ0  && CC0 &&  Ls1 &&  N2 &&  N3 && XOTHER;
    case CCALL:  return LIT1 && SZ0  && CC0 && 
                        (u->argc > 0                   ? TR1 : N1) && 
                        (u->argc > 1                   ? TR2 : N2) && 
                        (u->argc > 2 || u->has_ret_val ? TR3 : N3) &&
                        u->regparms_n <= u->argc && XCCALL;
-   /* Fields checked:     lit32   size  flags_r/w tag1   tag2   tag3    (rest) */
+   /* Fields checked:       lit32   size  flags_r/w tag1   tag2   tag3    (rest) */
    case MMX1:
-   case MMX2:        return LIT0 && SZ0  && CC0 &&  Ls1 &&  N2 &&  N3 && XOTHER;
-   case MMX3:        return LIT0 && SZ0  && CC0 &&  Ls1 && Ls2 &&  N3 && XOTHER;
-   case MMX2_MemRd:  return LIT0 && SZ48 && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
-   case MMX2_MemWr:  return LIT0 && SZ48 && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
-   case MMX2_ERegRd: return LIT0 && SZ4  && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
-   case MMX2_ERegWr: return LIT0 && SZ4  && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
+   case MMX2:         return LIT0 && SZ0  && CC0 &&  Ls1 &&  N2 &&  N3 && XOTHER;
+   case MMX3:         return LIT0 && SZ0  && CC0 &&  Ls1 && Ls2 &&  N3 && XOTHER;
+   case MMX2_MemRd:   return LIT0 && SZ48 && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
+   case MMX2_MemWr:   return LIT0 && SZ48 && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
+   case MMX2a1_MemRd: return LIT0 && SZ8  && CC0 &&  Ls1 && Ls2 && TR3 && XOTHER;
+   case MMX2_ERegRd:  return LIT0 && SZ4  && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
+   case MMX2_ERegWr:  return LIT0 && SZ4  && CC0 &&  Ls1 && TR2 &&  N3 && XOTHER;
 
    /* Fields checked:        lit32   size  flags_r/w tag1   tag2   tag3    (rest) */
    case SSE2a_MemWr:  return LIT0 && SZsse2 && CC0  && Ls1 && Ls2 && TR3 && XOTHER;
@@ -607,6 +653,7 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
 #  undef SZ42
 #  undef SZ48
 #  undef SZ416
+#  undef SZ816
 #  undef SZsse2
 #  undef SZsse3
 #  undef SZi
@@ -640,6 +687,8 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
 #  undef Ls3
 #  undef TRL1
 #  undef TRAL1
+#  undef TRA1
+#  undef TRA2
 #  undef N1
 #  undef N2
 #  undef N3
@@ -647,22 +696,35 @@ Bool VG_(saneUInstr) ( Bool beforeRA, Bool beforeLiveness, UInstr* u )
 #  undef Se1
 #  undef COND0
 #  undef EXTRA4b0
+#  undef EXTRA4b12
 #  undef SG_WD0
 #  undef JMPKIND0
 #  undef CCALL0
-#  undef Xextra4b
+#  undef XCONDi
+#  undef XLEA2
 #  undef XWIDEN
 #  undef XJMP
 #  undef XCCALL
 #  undef XOTHER
 }
 
-void VG_(saneUCodeBlock) ( UCodeBlock* cb )
+void VG_(sanity_check_UInstr)( UInt n, UInstr* u )
+{
+   Bool sane = is_sane_UInstr(u);
+   if (!sane) {
+      VG_(printf)("\nInsane instruction:\n");
+      VG_(pp_UInstr)(n, u);
+      VG_(up_UInstr)(n, u);
+      vg_assert(sane);
+   }
+}
+
+static void sanity_check_UCodeBlock ( UCodeBlock* cb )
 {
    Int i;
         
    for (i = 0; i < cb->used; i++) {
-      Bool sane = VG_(saneUInstr)(True, True, &cb->instrs[i]);
+      Bool sane = is_sane_UInstr(&cb->instrs[i]);
       if (!sane) {
          VG_(printf)("Instruction failed sanity check:\n");
          VG_(up_UInstr)(i, &cb->instrs[i]);
@@ -672,7 +734,7 @@ void VG_(saneUCodeBlock) ( UCodeBlock* cb )
 }
 
 /* Sanity checks to do with CALLMs in UCodeBlocks. */
-Bool VG_(saneUCodeBlockCalls) ( UCodeBlock* cb )
+static Bool is_sane_UCodeBlockCalls ( UCodeBlock* cb )
 {
    Int  callm = 0;
    Int  callm_s = 0;
@@ -753,6 +815,13 @@ Bool VG_(saneUCodeBlockCalls) ( UCodeBlock* cb )
    goto find_next_CALLM;
 }
 
+static void sanity_check_UCodeBlockCalls( UCodeBlock* cb )
+{
+   if ( ! is_sane_UCodeBlockCalls( cb ) ) {
+      VG_(pp_UCodeBlock)(cb, "block failing calls sanity check");
+      VG_(core_panic)("bad block");
+   }
+}
 
 /*------------------------------------------------------------*/
 /*--- Printing uinstrs.                                    ---*/
@@ -897,6 +966,7 @@ Char* VG_(name_UOpcode) ( Bool upper, Opcode opc )
       case MMX3:       return "MMX3" ;
       case MMX2_MemRd: return "MMX2_MRd" ;
       case MMX2_MemWr: return "MMX2_MWr" ;
+      case MMX2a1_MemRd: return "MMX2a1_MRd" ;
       case MMX2_ERegRd: return "MMX2_eRRd" ;
       case MMX2_ERegWr: return "MMX2_eRWr" ;
       case SSE2a_MemWr: return "SSE2a_MWr";
@@ -977,7 +1047,8 @@ void pp_UInstrWorker ( Int instrNo, UInstr* u, Bool ppRegsLiveness )
 {
    VG_(printf)("\t%4d: %s", instrNo, 
                             VG_(name_UOpcode)(True, u->opcode));
-   if (u->opcode == JMP || u->opcode == CC2VAL)
+   // For JMP, the condition goes before the size
+   if (u->opcode == JMP)
       VG_(printf)("%s", VG_(name_UCondcode)(u->cond));
 
    switch (u->size) {
@@ -990,6 +1061,27 @@ void pp_UInstrWorker ( Int instrNo, UInstr* u, Bool ppRegsLiveness )
       default: VG_(printf)("%d", (Int)u->size); break;
    }
 
+   // For CC2VAL and CMOV, the condition goes after the size
+   if (u->opcode == CC2VAL || u->opcode == CMOV)
+      VG_(printf)("%s", VG_(name_UCondcode)(u->cond));
+
+   // Append extra bits
+   switch (u->opcode) {
+   case JMP:
+      switch (u->jmpkind) {
+         case JmpCall:      VG_(printf)("-c"); break;
+         case JmpRet:       VG_(printf)("-r"); break;
+         case JmpSyscall:   VG_(printf)("-sys"); break;
+         case JmpClientReq: VG_(printf)("-cli"); break;
+         case JmpYield:     VG_(printf)("-yld"); break;
+         default: break;
+      }
+      break;
+
+   case WIDEN:
+      VG_(printf)("_%c%c", VG_(toupper)(nameISize(u->extra4b)),
+                           u->signed_widen?'s':'z');
+   } 
    VG_(printf)("       \t");
 
    switch (u->opcode) {
@@ -1067,6 +1159,12 @@ void pp_UInstrWorker ( Int instrNo, UInstr* u, Bool ppRegsLiveness )
          VG_(pp_UOperand)(u, 2, 4, True);
          break;
 
+      case MMX2a1_MemRd:
+          VG_(printf)("0x%x:0x%x:0x%x",
+                     (u->val1 >> 8) & 0xFF, u->val1 & 0xFF, u->val2 & 0xFF );
+         VG_(pp_UOperand)(u, 3, 4, True);
+         break;
+
       case SSE2a_MemWr:
       case SSE2a_MemRd:
       case SSE2g_RegWr:
@@ -1140,14 +1238,6 @@ void pp_UInstrWorker ( Int instrNo, UInstr* u, Bool ppRegsLiveness )
          break;
 
       case JMP:
-         switch (u->jmpkind) {
-            case JmpCall:      VG_(printf)("-c"); break;
-            case JmpRet:       VG_(printf)("-r"); break;
-            case JmpSyscall:   VG_(printf)("-sys"); break;
-            case JmpClientReq: VG_(printf)("-cli"); break;
-            case JmpYield:     VG_(printf)("-yld"); break;
-            default: break;
-         }
          VG_(pp_UOperand)(u, 1, u->size, False);
          if (CondAlways == u->cond) {
             /* Print x86 instruction size if filled in */
@@ -1202,8 +1292,6 @@ void pp_UInstrWorker ( Int instrNo, UInstr* u, Bool ppRegsLiveness )
          break;
 
       case WIDEN:
-         VG_(printf)("_%c%c", VG_(toupper)(nameISize(u->extra4b)),
-                              u->signed_widen?'s':'z');
          VG_(pp_UOperand)(u, 1, u->size, False);
          break;
 
@@ -1296,6 +1384,7 @@ Int VG_(get_reg_usage) ( UInstr* u, Tag tag, Int* regs, Bool* isWrites )
 
       case SSE3ag_MemRd_RegWr: RD(1); WR(2); break;
 
+      case MMX2a1_MemRd: RD(3); break;
       case MMX2_ERegRd: RD(2); break;
       case MMX2_ERegWr: WR(2); break;
 
@@ -1415,7 +1504,7 @@ Int containingArchRegOf ( Int sz, Int aregno )
 
 /* If u reads an ArchReg, return the number of the containing arch
    reg.  Otherwise return -1.  Used in redundant-PUT elimination.
-   Note that this is not required for skins extending UCode because
+   Note that this is not required for tools extending UCode because
    this happens before instrumentation. */
 static
 Int maybe_uinstrReadsArchReg ( UInstr* u )
@@ -1451,7 +1540,7 @@ Int maybe_uinstrReadsArchReg ( UInstr* u )
       case JIFZ:
       case FPU: case FPU_R: case FPU_W:
       case MMX1: case MMX2: case MMX3:
-      case MMX2_MemRd: case MMX2_MemWr:
+      case MMX2_MemRd: case MMX2_MemWr: case MMX2a1_MemRd:
       case MMX2_ERegRd: case MMX2_ERegWr:
       case SSE2a_MemWr: case SSE2a_MemRd: case SSE2a1_MemRd:
       case SSE2g_RegWr: case SSE2g1_RegWr: case SSE2e1_RegRd:
@@ -1802,9 +1891,9 @@ static void vg_improve ( UCodeBlock* cb )
 /*--- %ESP-update pass                                     ---*/
 /*------------------------------------------------------------*/
 
-/* For skins that want to know about %ESP changes, this pass adds
-   in the appropriate hooks.  We have to do it after the skin's
-   instrumentation, so the skin doesn't have to worry about the CCALLs
+/* For tools that want to know about %ESP changes, this pass adds
+   in the appropriate hooks.  We have to do it after the tool's
+   instrumentation, so the tool doesn't have to worry about the CCALLs
    it adds in, and we must do it before register allocation because
    spilled temps make it much harder to work out the %esp deltas.
    Thus we have it as an extra phase between the two. 
@@ -1920,7 +2009,7 @@ typedef
    struct {
       /* Becomes live for the first time after this insn ... */
       Int live_after;
-      /* Becomes dead for the last time after this insn ... */
+      /* Becomes dead for the last time before this insn ... */
       Int dead_before;
       /* The "home" spill slot, if needed.  Never changes. */
       Int spill_no;
@@ -2009,7 +2098,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
             if (temp_info[tno].live_after == VG_NOTHING) {
                VG_(printf)("At instr %d...\n", i);
                VG_(core_panic)("First use of tmp not a write,"
-                               " probably a skin instrumentation error");
+                               " probably a tool instrumentation error");
             }
             /* Reads only hold it live until before this insn. */
             if (temp_info[tno].dead_before < i)
@@ -2059,7 +2148,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
          max_ss_no = j;
    }
 
-   VG_(total_reg_rank) += (max_ss_no+1);
+   n_total_reg_rank += (max_ss_no+1);
 
    /* Show live ranges and assigned spill slot nos. */
 
@@ -2089,8 +2178,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
 
    /* Resulting code goes here.  We generate it all in a forwards
       pass. */
-   c2 = VG_(alloc_UCodeBlock)();
-   c2->orig_eip = c1->orig_eip;
+   c2 = alloc_UCodeBlock( c1->orig_eip );
 
    /* At the start, no TempRegs are assigned to any real register.
       Correspondingly, all temps claim to be currently resident in
@@ -2104,7 +2192,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
    for (i = 0; i < c1->used; i++) {
 
       if (c1->instrs[i].opcode == NOP) continue;
-      VG_(uinstrs_prealloc)++;
+      n_uinstrs_prealloc++;
 
 #     if 0
       /* Check map consistency.  Expensive but correct. */
@@ -2240,7 +2328,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
             uInstr2(c2, PUT, 4, 
                         RealReg, VG_(rank_to_realreg)(r), 
                         SpillNo, temp_info[real_to_temp[r]].spill_no);
-            VG_(uinstrs_spill)++;
+            n_uinstrs_spill++;
             spill_reqd = True;
             if (dis)
                VG_(pp_UInstr)(c2->used-1, &LAST_UINSTR(c2));
@@ -2257,7 +2345,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
             uInstr2(c2, GET, 4, 
                         SpillNo, temp_info[tno].spill_no, 
                         RealReg, VG_(rank_to_realreg)(r) );
-            VG_(uinstrs_spill)++;
+            n_uinstrs_spill++;
             spill_reqd = True;
             if (dis)
                VG_(pp_UInstr)(c2->used-1, &LAST_UINSTR(c2));
@@ -2289,7 +2377,7 @@ UCodeBlock* vg_do_register_allocation ( UCodeBlock* c1 )
    VG_(free_UCodeBlock)(c1);
 
    if (spill_reqd) 
-      VG_(translations_needing_spill)++;
+      n_translations_needing_spill++;
 
    return c2;
 
@@ -2332,45 +2420,43 @@ static void vg_realreg_liveness_analysis ( UCodeBlock* cb )
 /*--- Main entry point for the JITter.                     ---*/
 /*------------------------------------------------------------*/
 
-/* Translate the basic block beginning at orig_addr, placing the
-   translation in a vg_malloc'd block, the address and size of which
-   are returned in trans_addr and trans_size.  Length of the original
-   block is also returned in orig_size.  If the latter three are NULL,
-   this call is being done for debugging purposes, in which case (a)
-   throw away the translation once it is made, and (b) produce a load
-   of debugging output. 
+/* Translate the basic block beginning at orig_addr, and add it to
+   the translation cache & translation table.  Unless 'debugging' is true,
+   in which case the call is being done for debugging purposes, so
+   (a) throw away the translation once it is made, and (b) produce a
+   load of debugging output. 
 
-   'tst' is the identity of the thread needing this block.
+   'tid' is the identity of the thread needing this block.
 */
-void VG_(translate) ( /*IN*/  ThreadId tid, 
-		      /*IN*/  Addr  orig_addr,  
-                      /*OUT*/ UInt* orig_size,
-                      /*OUT*/ Addr* trans_addr, 
-                      /*OUT*/ UInt* trans_size,
-		      /*OUT*/ UShort jumps[VG_MAX_JUMPS])
+void VG_(translate) ( ThreadId tid, Addr orig_addr,
+                      Bool debugging_translation )
 {
-   Int         n_disassembled_bytes, final_code_size;
-   Bool        debugging_translation;
-   UChar*      final_code;
+   Addr        trans_addr, redir, orig_addr0 = orig_addr;
+   UShort      jumps[VG_MAX_JUMPS];
+   Int         i, orig_size, trans_size;
    UCodeBlock* cb;
    Bool        notrace_until_done;
    UInt        notrace_until_limit = 0;
    Segment     *seg;
-   Addr		redir;
 
    VGP_PUSHCC(VgpTranslate);
-   debugging_translation
-      = orig_size == NULL || trans_addr == NULL || trans_size == NULL;
+
+   beforeRA        = True;
+   beforeLiveness  = True;
+
+   for (i = 0; i < VG_MAX_JUMPS; i++)
+      jumps[i] = (UShort)-1;
 
    /* Look in the code redirect table to see if we should
       translate an alternative address for orig_addr. */
    redir = VG_(code_redirect)(orig_addr);
 
-   if (redir != orig_addr && VG_(clo_verbosity) >= 2)
+   if (redir != orig_addr && VG_(clo_verbosity) >= 2) {
       VG_(message)(Vg_UserMsg, 
 		   "TRANSLATE: %p redirected to %p",
 		   orig_addr, 
 		   redir );
+   }
    orig_addr = redir;
 
    /* If codegen tracing, don't start tracing until
@@ -2379,8 +2465,7 @@ void VG_(translate) ( /*IN*/  ThreadId tid,
       few blocks translated prior to a failure.  Set
       notrace_until_limit to be the number of translations to be made
       before --trace-codegen= style printing takes effect. */
-   notrace_until_done
-      = VG_(overall_in_count) >= notrace_until_limit;
+   notrace_until_done = VG_(get_bbs_translated)() >= notrace_until_limit;
 
    seg = VG_(find_segment)(orig_addr);
 
@@ -2390,36 +2475,28 @@ void VG_(translate) ( /*IN*/  ThreadId tid,
    if (seg == NULL ||
        !VG_(seg_contains)(seg, orig_addr, 1) || 
        (seg->prot & (VKI_PROT_READ|VKI_PROT_EXEC)) == 0) {
-      vki_ksiginfo_t info;
-
       /* Code address is bad - deliver a signal instead */
       vg_assert(!VG_(is_addressable)(orig_addr, 1));
 
-      info.si_signo = VKI_SIGSEGV;
-
       if (seg != NULL && VG_(seg_contains)(seg, orig_addr, 1)) {
 	 vg_assert((seg->prot & VKI_PROT_EXEC) == 0);
-	 info.si_code = 2;	/* invalid permissions for mapped object */
+	 VG_(synth_fault_perms)(tid, orig_addr);
       } else
-	 info.si_code = 1;	/* address not mapped to object */
-      info._sifields._sigfault._addr = (void*)orig_addr;
+	 VG_(synth_fault_mapping)(tid, orig_addr);
 
-      VG_(deliver_signal)(tid, &info, False);
       return;
    } else
       seg->flags |= SF_CODE;	/* contains cached code */
 
-   cb = VG_(alloc_UCodeBlock)();
-   cb->orig_eip = orig_addr;
+   cb = alloc_UCodeBlock( orig_addr );
 
    /* If doing any code printing, print a basic block start marker */
    if (VG_(clo_trace_codegen) && notrace_until_done) {
       Char fnname[64] = "";
       VG_(get_fnname_if_entry)(orig_addr, fnname, 64);
       VG_(printf)(
-              "==== BB %d %s(%p) in %dB, out %dB, BBs exec'd %llu ====\n\n",
-              VG_(overall_in_count), fnname, orig_addr, 
-              VG_(overall_in_osize), VG_(overall_in_tsize),
+              "==== BB %d %s(%p) approx BBs exec'd %llu ====\n\n",
+              VG_(get_bbs_translated)(), fnname, orig_addr, 
               VG_(bbs_done));
    }
 
@@ -2432,7 +2509,11 @@ void VG_(translate) ( /*IN*/  ThreadId tid,
    /* Disassemble this basic block into cb. */
    VG_(print_codegen) = DECIDE_IF_PRINTING_CODEGEN_FOR_PHASE(1);
    VGP_PUSHCC(VgpToUCode);
-   n_disassembled_bytes = VG_(disBB) ( cb, orig_addr );
+   orig_size = VG_(disBB) ( cb, orig_addr );
+   sanity_check_UCodeBlock ( cb );
+   // Only sanity-check calls now because tools might remove the
+   // CALLM_[ES] pairs.
+   sanity_check_UCodeBlockCalls ( cb );
    VGP_POPCC(VgpToUCode);
 
    /* Try and improve the code a bit. */
@@ -2450,10 +2531,10 @@ void VG_(translate) ( /*IN*/  ThreadId tid,
    cb = SK_(instrument) ( cb, orig_addr );
    if (VG_(print_codegen))
       VG_(pp_UCodeBlock) ( cb, "Instrumented UCode:" );
-   VG_(saneUCodeBlock)( cb );
+   sanity_check_UCodeBlock( cb );
    VGP_POPCC(VgpInstrument);
 
-   /* Add %ESP-update hooks if the skin requires them */
+   /* Add %ESP-update hooks if the tool requires them */
    /* Nb: We don't print out this phase, because it doesn't do much */
    if (VG_(need_to_handle_esp_assignment)()) {
       VGP_PUSHCC(VgpESPUpdate);
@@ -2465,36 +2546,46 @@ void VG_(translate) ( /*IN*/  ThreadId tid,
    VG_(print_codegen) = DECIDE_IF_PRINTING_CODEGEN_FOR_PHASE(4);
    VGP_PUSHCC(VgpRegAlloc);
    cb = vg_do_register_allocation ( cb );
+   beforeRA = False;
    VGP_POPCC(VgpRegAlloc);
 
    /* Do post reg-alloc %e[acd]x liveness analysis (too boring to print
     * anything;  results can be seen when emitting final code). */
    VGP_PUSHCC(VgpLiveness);
    vg_realreg_liveness_analysis ( cb );
+   beforeLiveness = False;
    VGP_POPCC(VgpLiveness);
 
    /* Emit final code */
    VG_(print_codegen) = DECIDE_IF_PRINTING_CODEGEN_FOR_PHASE(5);
-
    VGP_PUSHCC(VgpFromUcode);
-   final_code = VG_(emit_code)(cb, &final_code_size, jumps );
+   trans_addr = (Addr)VG_(emit_code)(cb, &trans_size, jumps );
    VGP_POPCC(VgpFromUcode);
    VG_(free_UCodeBlock)(cb);
 
 #undef DECIDE_IF_PRINTING_CODEGEN_FOR_PHASE
 
-   if (debugging_translation) {
-      /* Only done for debugging -- throw away final result. */
-      VG_(arena_free)(VG_AR_JITTER, final_code);
-   } else {
-      /* Doing it for real -- return values to caller. */
-      *orig_size = n_disassembled_bytes;
-      *trans_addr = (Addr)final_code;
-      *trans_size = final_code_size;
+   /* Copy data at trans_addr into the translation cache. */
+   /* Since the .orig_size and .trans_size fields are UShort, be paranoid. */
+   vg_assert(orig_size  > 0 && orig_size  < 65536);
+   vg_assert(trans_size > 0 && trans_size < 65536);
+
+   // If debugging, don't do anything with the translated block;  we
+   // only did this for the debugging output produced along the way.
+   if (!debugging_translation) {
+      // Note that we use orig_addr0, not orig_addr, which might have been
+      // changed by the redirection
+      VG_(add_to_trans_tab)( orig_addr0, orig_size, trans_addr, trans_size,
+                             jumps );
    }
+
+   /* Free the intermediary -- was allocated by VG_(emit_code). */
+   VG_(arena_free)( VG_AR_JITTER, (void*)trans_addr );
+
    VGP_POPCC(VgpTranslate);
 }
 
+
 /*--------------------------------------------------------------------*/
 /*--- end                                           vg_translate.c ---*/
 /*--------------------------------------------------------------------*/
