@@ -12,7 +12,7 @@
    This file is part of Valgrind, an extensible x86 protected-mode
    emulator for monitoring program execution on x86-Unixes.
 
-   Copyright (C) 2000-2003 Julian Seward.  All rights reserved.
+   Copyright (C) 2000-2004 Julian Seward.  All rights reserved.
 
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
@@ -159,20 +159,28 @@ typedef
              Valgrind's output to /dev/null and still count errors. */
           VG_USERREQ__COUNT_ERRORS = 0x1201,
 
-          /* These are useful and can be interpreted by any skin that tracks
+          /* These are useful and can be interpreted by any tool that tracks
              malloc() et al, by using vg_replace_malloc.c. */
           VG_USERREQ__MALLOCLIKE_BLOCK = 0x1301,
           VG_USERREQ__FREELIKE_BLOCK   = 0x1302,
+          /* Memory pool support. */
+          VG_USERREQ__CREATE_MEMPOOL   = 0x1303,
+          VG_USERREQ__DESTROY_MEMPOOL  = 0x1304,
+          VG_USERREQ__MEMPOOL_ALLOC    = 0x1305,
+          VG_USERREQ__MEMPOOL_FREE     = 0x1306,
 
           /* Allow printfs to valgrind log. */
           VG_USERREQ__PRINTF = 0x1401,
-          VG_USERREQ__PRINTF_BACKTRACE = 0x1402,
+          VG_USERREQ__PRINTF_BACKTRACE = 0x1402
    } Vg_ClientRequest;
 
+#ifndef __GNUC__
+#define __extension__
+#endif
 
 /* Returns 1 if running on Valgrind, 0 if running on the real CPU. 
    Currently implemented but untested. */
-#define RUNNING_ON_VALGRIND                                        \
+#define RUNNING_ON_VALGRIND  __extension__                         \
    ({unsigned int _qzz_res;                                        \
     VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0 /* returned if not */,     \
                             VG_USERREQ__RUNNING_ON_VALGRIND,       \
@@ -270,8 +278,8 @@ VALGRIND_PRINTF_BACKTRACE(const char *format, ...)
    })
 
 
-/* Counts the number of errors that have been recorded by a skin.  Nb:
-   the skin must record the errors with VG_(maybe_record_error)() or
+/* Counts the number of errors that have been recorded by a tool.  Nb:
+   the tool must record the errors with VG_(maybe_record_error)() or
    VG_(unique_error)() for them to be counted. */
 #define VALGRIND_COUNT_ERRORS                                           \
    ({unsigned int _qyy_res;                                             \
@@ -319,4 +327,36 @@ VALGRIND_PRINTF_BACKTRACE(const char *format, ...)
                             addr, rzB, 0, 0);                      \
    }
 
+/* Create a memory pool. */
+#define VALGRIND_CREATE_MEMPOOL(pool, rzB, is_zeroed)              \
+   {unsigned int _qzz_res;                                         \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__CREATE_MEMPOOL,            \
+                            pool, rzB, is_zeroed, 0);              \
+   }
+
+/* Destroy a memory pool. */
+#define VALGRIND_DESTROY_MEMPOOL(pool)                             \
+   {unsigned int _qzz_res;                                         \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__DESTROY_MEMPOOL,           \
+                            pool, 0, 0, 0);                        \
+   }
+
+/* Associate a piece of memory with a memory pool. */
+#define VALGRIND_MEMPOOL_ALLOC(pool, addr, size)                   \
+   {unsigned int _qzz_res;                                         \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__MEMPOOL_ALLOC,             \
+                            pool, addr, size, 0);                  \
+   }
+
+/* Disassociate a piece of memory from a memory pool. */
+#define VALGRIND_MEMPOOL_FREE(pool, addr)                          \
+   {unsigned int _qzz_res;                                         \
+    VALGRIND_MAGIC_SEQUENCE(_qzz_res, 0,                           \
+                            VG_USERREQ__MEMPOOL_FREE,              \
+                            pool, addr, 0, 0);                     \
+   }
+
 #endif   /* __VALGRIND_H */
