@@ -249,9 +249,9 @@ struct vki_ucontext {
 #define VKI_SA_NODEFER      0x40000000
 #define VKI_SA_NOMASK       VKI_SA_NODEFER
 #define VKI_SA_NOCLDWAIT    0x00000002
+#define VKI_SA_RESTORER     0x04000000
 #if 0
 #define VKI_SA_INTERRUPT    0x20000000 /* dummy -- ignored */
-#define VKI_SA_RESTORER     0x04000000
 #endif
 
 /* extra wait flags */
@@ -546,6 +546,12 @@ struct vki_timeval {
 #define VKI_F_SETFD	    2	    /* set/clear close_on_exec */
 #define VKI_F_GETFL         3       /* get file->f_flags */
 #define VKI_F_SETFL         4       /* set file->f_flags */
+#define VKI_F_GETLK         5
+#define VKI_F_SETLK         6
+#define VKI_F_SETLKW        7
+#define VKI_F_GETLK64       12      /*  using 'struct flock64' */
+#define VKI_F_SETLK64       13
+#define VKI_F_SETLKW64      14
 
 /* for F_[GET|SET]FL */
 #define VKI_FD_CLOEXEC	    1	    /* actually anything with low bit set goes */
@@ -559,6 +565,11 @@ struct vki_timespec {
         long    tv_nsec;        /* nanoseconds */
 };
 
+/* POSIX.1b structure for timer start values and intervals.  */
+struct vki_itimerspec {
+	struct vki_timespec it_interval;
+	struct vki_timespec it_value;
+};
 
 /* STAT stuff 
    from /usr/src/linux-2.4.9-31/include/asm-i386/stat.h */
@@ -726,7 +737,7 @@ struct vki_sockaddr {
  * bits/statfs.h
  */
 
-struct statfs {
+struct vki_statfs {
         unsigned int f_type;
         unsigned int f_bsize;
         unsigned int f_blocks;
@@ -740,7 +751,7 @@ struct statfs {
         unsigned int f_spare[5];
 };
 
-struct statfs64 {
+struct vki_statfs64 {
         unsigned int f_type;
         unsigned int f_bsize;
         unsigned long long f_blocks;
@@ -763,6 +774,201 @@ struct statfs64 {
 #define VKI_FUTEX_FD      2
 #define VKI_FUTEX_REQUEUE 3
 
+/*
+ * linux/ipc.h
+ */
+
+#define VKI_IPC_CREAT  00001000   /* create if key is nonexistent */
+#define VKI_IPC_EXCL   00002000   /* fail if key exists */
+#define VKI_IPC_NOWAIT 00004000   /* return error on wait */
+
+/* 
+ * linux/elfcore.h
+ */
+
+struct elf_siginfo
+{
+	int	si_signo;			/* signal number */
+	int	si_code;			/* extra code */
+	int	si_errno;			/* errno */
+};
+
+/*
+ * This is the old layout of "struct pt_regs", and
+ * is still the layout used by user mode (the new
+ * pt_regs doesn't have all registers as the kernel
+ * doesn't use the extra segment registers)
+ */
+struct user_regs_struct {
+	long ebx, ecx, edx, esi, edi, ebp, eax;
+	unsigned short ds, __ds, es, __es;
+	unsigned short fs, __fs, gs, __gs;
+	long orig_eax, eip;
+	unsigned short cs, __cs;
+	long eflags, esp;
+	unsigned short ss, __ss;
+};
+
+struct user_i387_struct {
+	long	cwd;
+	long	swd;
+	long	twd;
+	long	fip;
+	long	fcs;
+	long	foo;
+	long	fos;
+	long	st_space[20];	/* 8*10 bytes for each FP-reg = 80 bytes */
+};
+
+struct user_fxsr_struct {
+	unsigned short	cwd;
+	unsigned short	swd;
+	unsigned short	twd;
+	unsigned short	fop;
+	long	fip;
+	long	fcs;
+	long	foo;
+	long	fos;
+	long	mxcsr;
+	long	reserved;
+	long	st_space[32];	/* 8*16 bytes for each FP-reg = 128 bytes */
+	long	xmm_space[32];	/* 8*16 bytes for each XMM-reg = 128 bytes */
+	long	padding[56];
+};
+
+typedef unsigned long elf_greg_t;
+
+#define ELF_NGREG (sizeof (struct user_regs_struct) / sizeof(elf_greg_t))
+typedef elf_greg_t elf_gregset_t[ELF_NGREG];
+
+typedef struct user_i387_struct elf_fpregset_t;
+typedef struct user_fxsr_struct elf_fpxregset_t;
+
+
+/*
+ * Definitions to generate Intel SVR4-like core files.
+ * These mostly have the same names as the SVR4 types with "elf_"
+ * tacked on the front to prevent clashes with linux definitions,
+ * and the typedef forms have been avoided.  This is mostly like
+ * the SVR4 structure, but more Linuxy, with things that Linux does
+ * not support and which gdb doesn't really use excluded.
+ * Fields present but not used are marked with "XXX".
+ */
+struct elf_prstatus
+{
+#if 0
+	long	pr_flags;	/* XXX Process flags */
+	short	pr_why;		/* XXX Reason for process halt */
+	short	pr_what;	/* XXX More detailed reason */
+#endif
+	struct elf_siginfo pr_info;	/* Info associated with signal */
+	short	pr_cursig;		/* Current signal */
+	unsigned long pr_sigpend;	/* Set of pending signals */
+	unsigned long pr_sighold;	/* Set of held signals */
+#if 0
+	struct sigaltstack pr_altstack;	/* Alternate stack info */
+	struct sigaction pr_action;	/* Signal action for current sig */
+#endif
+	Int	pr_pid;
+	Int	pr_ppid;
+	Int	pr_pgrp;
+	Int	pr_sid;
+	struct vki_timeval pr_utime;	/* User time */
+	struct vki_timeval pr_stime;	/* System time */
+	struct vki_timeval pr_cutime;	/* Cumulative user time */
+	struct vki_timeval pr_cstime;	/* Cumulative system time */
+#if 0
+	long	pr_instr;		/* Current instruction */
+#endif
+	elf_gregset_t pr_reg;	/* GP registers */
+	int pr_fpvalid;		/* True if math co-processor being used.  */
+};
+
+#define ELF_PRARGSZ	(80)	/* Number of chars for args */
+
+struct elf_prpsinfo
+{
+	char	pr_state;	/* numeric process state */
+	char	pr_sname;	/* char for pr_state */
+	char	pr_zomb;	/* zombie */
+	char	pr_nice;	/* nice val */
+	unsigned long pr_flag;	/* flags */
+	Int	pr_uid;
+	Int	pr_gid;
+	Int	pr_pid, pr_ppid, pr_pgrp, pr_sid;
+	/* Lots missing */
+	char	pr_fname[16];	/* filename of executable */
+	char	pr_psargs[ELF_PRARGSZ];	/* initial part of arg list */
+};
+
+/*
+ * linux/aio_abi.h
+ */
+
+typedef struct {
+	unsigned        id;     /* kernel internal index number */
+	unsigned        nr;     /* number of io_events */
+	unsigned        head;
+	unsigned        tail;
+ 
+	unsigned        magic;
+	unsigned        compat_features;
+	unsigned        incompat_features;
+	unsigned        header_length;  /* size of aio_ring */  
+} vki_aio_ring ;
+
+typedef vki_aio_ring *vki_aio_context_t;
+
+typedef struct {
+    ULong data;
+    ULong obj;
+    Long  result;
+    Long  result2;
+} vki_io_event;
+
+typedef struct {
+	/* these are internal to the kernel/libc. */
+	ULong	aio_data;	/* data to be returned in event's data */
+        ULong	aio_key;
+				/* the kernel sets aio_key to the req # */
+
+	/* common fields */
+        UShort	aio_lio_opcode;	/* see IOCB_CMD_ above */
+	UShort	aio_reqprio;
+	UInt	aio_fildes;
+
+	ULong	aio_buf;
+	ULong	aio_nbytes;
+	Long	aio_offset;
+
+	/* extra parameters */
+	ULong	aio_reserved2;	/* TODO: use this for a (struct sigevent *) */
+	ULong	aio_reserved3;
+} vki_iocb; /* 64 bytes */
+
+enum {
+    VKI_IOCB_CMD_PREAD = 0,
+    VKI_IOCB_CMD_PWRITE = 1,
+    VKI_IOCB_CMD_FSYNC = 2,
+    VKI_IOCB_CMD_FDSYNC = 3,
+    /* These two are experimental.
+     * IOCB_CMD_PREADX = 4,
+     * IOCB_CMD_POLL = 5,
+     */
+    VKI_IOCB_CMD_NOOP = 6,
+};
+
+/*
+ * linux/mqueue.h
+ */
+
+struct vki_mq_attr {
+	long	mq_flags;	/* message queue flags			*/
+	long	mq_maxmsg;	/* maximum number of messages		*/
+	long	mq_msgsize;	/* maximum message size			*/
+	long	mq_curmsgs;	/* number of messages currently queued	*/
+	long	__reserved[4];	/* ignored for input, zeroed for output */
+};
 
 #endif /*  __VG_KERNELIFACE_H */
 
