@@ -9,7 +9,7 @@
    detecting memory errors, and AddrCheck, a lightweight Valgrind tool 
    for detecting memory errors.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -64,8 +64,8 @@ void vg_scan_all_valid_memory_sighandler ( Int sigNo )
    on the page is valid.  Only if so are that address and its contents
    passed to notify_word.
 
-   This is all to avoid duplication of this machinery between the
-   memcheck and addrcheck skins.  
+   This is all to avoid duplication of this machinery between
+   Memcheck and Addrcheck.  
 */
 static
 UInt vg_scan_all_valid_memory ( Bool is_valid_64k_chunk ( UInt ),
@@ -311,10 +311,7 @@ void vg_detect_memory_leaks_notify_addr ( Addr a, UInt word_at_a )
       where the .bss segment has been put.  If you can, drop me a
       line.  
    */
-   if (VG_(within_stack)(a))                      return;
-   if (VG_(within_m_state_static_OR_threads)(a))  return;
-   if (a == (Addr)(&lc_min_mallocd_addr))         return;
-   if (a == (Addr)(&lc_max_mallocd_addr))         return;
+   if (!VG_(is_client_addr)(a))			  return;
 
    /* OK, let's get on and do something Useful for a change. */
 
@@ -371,8 +368,8 @@ static Int lc_compar(void* n1, void* n2)
 /* Top level entry point to leak detector.  Call here, passing in
    suitable address-validating functions (see comment at top of
    vg_scan_all_valid_memory above).  All this is to avoid duplication
-   of the leak-detection code for the Memcheck and Addrcheck skins.
-   Also pass in a skin-specific function to extract the .where field
+   of the leak-detection code for Memcheck and Addrcheck.
+   Also pass in a tool-specific function to extract the .where field
    for allocated blocks, an indication of the resolution wanted for
    distinguishing different allocation points, and whether or not
    reachable blocks should be shown.
