@@ -9,7 +9,7 @@
    This file is part of MemCheck, a heavyweight Valgrind tool for
    detecting memory errors.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -31,8 +31,26 @@
 */
 
 #include "mc_include.h"
+#include "memcheck.h"
 #include "valgrind.h"
 
+static Addr record_overlap_error;
+
+static int init_done;
+
+/* Startup hook - called as init section */
+static void init(void) __attribute__((constructor));
+static void init(void) 
+{
+   if (init_done)
+      return;
+
+   VALGRIND_MAGIC_SEQUENCE(record_overlap_error, 0,
+			   _VG_USERREQ__MEMCHECK_GET_RECORD_OVERLAP,
+			   0, 0, 0, 0);
+   init_done = 1;
+}
+
 /* ---------------------------------------------------------------------
    The normal versions of these functions are hyper-optimised, which fools
    Memcheck and cause spurious value warnings.  So we replace them with
@@ -80,7 +98,8 @@ void complain2 ( Char* s, char* dst, const char* src )
    OverlapExtra extra = {
       .src = (Addr)src, .dst = (Addr)dst, .len = -1,
    };
-   VALGRIND_NON_SIMD_CALL2( MAC_(record_overlap_error), s, &extra );
+   init();
+   VALGRIND_NON_SIMD_CALL2( record_overlap_error, s, &extra );
 }
 
 static __inline__
@@ -90,7 +109,8 @@ void complain3 ( Char* s, void* dst, const void* src, int n )
    OverlapExtra extra = {
       .src = (Addr)src, .dst = (Addr)dst, .len = n,
    };
-   VALGRIND_NON_SIMD_CALL2( MAC_(record_overlap_error), s, &extra );
+   init();
+   VALGRIND_NON_SIMD_CALL2( record_overlap_error, s, &extra );
 }
 
 char* strrchr ( const char* s, int c )
@@ -135,11 +155,11 @@ char* strcat ( char* dst, const char* src )
    return dst_orig;
 }
 
-char* strncat ( char* dst, const char* src, int n )
+char* strncat ( char* dst, const char* src, unsigned int n )
 {
    const Char* src_orig = src;
          Char* dst_orig = dst;
-   Int   m = 0;
+   UInt  m = 0;
 
    while (*dst) dst++;
    while (m   < n && *src) { m++; *dst++ = *src++; } /* concat <= n chars */
@@ -156,6 +176,13 @@ char* strncat ( char* dst, const char* src, int n )
    return dst_orig;
 }
 
+unsigned int strnlen ( const char* str, unsigned int n )
+{
+   UInt i = 0;
+   while (i < n && str[i] != 0) i++;
+   return i;
+}
+
 unsigned int strlen ( const char* str )
 {
    UInt i = 0;
@@ -182,11 +209,11 @@ char* strcpy ( char* dst, const char* src )
    return dst_orig;
 }
 
-char* strncpy ( char* dst, const char* src, int n )
+char* strncpy ( char* dst, const char* src, unsigned int n )
 {
    const Char* src_orig = src;
          Char* dst_orig = dst;
-   Int   m = 0;
+   UInt  m = 0;
 
    while (m   < n && *src) { m++; *dst++ = *src++; }
    /* Check for overlap after copying; all n bytes of dst are relevant,
@@ -198,8 +225,7 @@ char* strncpy ( char* dst, const char* src, int n )
    return dst_orig;
 }
 
-int strncmp ( const unsigned char* s1, const unsigned char* s2, 
-              unsigned int nmax )
+int strncmp ( const char* s1, const char* s2, unsigned int nmax )
 {
    unsigned int n = 0;
    while (True) {
@@ -246,9 +272,12 @@ void* memcpy( void *dst, const void *src, unsigned int len )
    register char *d;
    register char *s;
 
+   if (len == 0)
+      return dst;
+
    if (is_overlap(dst, src, len, len))
       complain3("memcpy", dst, src, len);
-      
+
    if ( dst > src ) {
       d = (char *)dst + len - 1;
       s = (char *)src + len - 1;
@@ -300,6 +329,30 @@ int memcmp ( const void *s1V, const void *s2V, unsigned int n )
    return 0;
 }
 
+
+/* Copy SRC to DEST, returning the address of the terminating '\0' in
+   DEST. (minor variant of strcpy) */
+
+char* stpcpy ( char* dst, const char* src )
+{
+   const Char* src_orig = src;
+         Char* dst_orig = dst;
+
+   while (*src) *dst++ = *src++;
+   *dst = 0;
+
+   /* This checks for overlap after copying, unavoidable without
+      pre-counting length... should be ok */
+   if (is_overlap(dst_orig, 
+                  src_orig, 
+                  (Addr)dst-(Addr)dst_orig+1, 
+                  (Addr)src-(Addr)src_orig+1))
+      complain2("stpcpy", dst_orig, src_orig);
+
+   return dst;
+}
+
+
 /*--------------------------------------------------------------------*/
 /*--- end                                     mac_replace_strmem.c ---*/
 /*--------------------------------------------------------------------*/
