@@ -9,7 +9,7 @@
    detecting memory errors, and AddrCheck, a lightweight Valgrind tool 
    for detecting memory errors.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -51,44 +51,38 @@ typedef
       Stack, 
       Unknown,      /* classification yielded nothing useful */
       Freed, Mallocd, 
-      UserG  /* in a user-defined block; Addrcheck & Memcheck only */
+      UserG,        /* in a user-defined block; Addrcheck & Memcheck only */
+      Mempool,      /* in a mempool; Addrcheck & Memcheck only */
    }
    AddrKind;
 
 /* Records info about a faulting address. */
 typedef
-   struct {
-      /* ALL */
-      AddrKind akind;
-      /* Freed, Mallocd */
-      Int blksize;
-      /* Freed, Mallocd */
-      Int rwoffset;
-      /* Freed, Mallocd */
-      ExeContext* lastchange;
-      /* Stack */
-      ThreadId stack_tid;
-      /* True if is just-below %esp -- could be a gcc bug. */
-      Bool maybe_gcc;
+   struct {                   // Used by:
+      AddrKind akind;         //   ALL
+      Int blksize;            //   Freed, Mallocd
+      Int rwoffset;           //   Freed, Mallocd
+      ExeContext* lastchange; //   Freed, Mallocd
+      ThreadId stack_tid;     //   Stack
+      Bool maybe_gcc;         // True if just below %esp -- could be a gcc bug.
    }
    AddrInfo;
 
 typedef 
    enum { 
-      /* Bad syscall params */
-      ParamSupp,
-      /* Memory errors in core (pthread ops, signal handling) */
-      CoreMemSupp,
-      /* Use of invalid values of given size (MemCheck only) */
+      ParamSupp,     // Bad syscall params
+      CoreMemSupp,   // Memory errors in core (pthread ops, signal handling)
+
+      // Use of invalid values of given size (MemCheck only)
       Value0Supp, Value1Supp, Value2Supp, Value4Supp, Value8Supp, Value16Supp,
-      /* Invalid read/write attempt at given size */
+
+      // Invalid read/write attempt at given size
       Addr1Supp, Addr2Supp, Addr4Supp, Addr8Supp, Addr16Supp,
-      /* Invalid or mismatching free */
-      FreeSupp,
-      /* Overlapping blocks in memcpy(), strcpy(), etc */
-      OverlapSupp,
-      /* Something to be suppressed in a leak check. */
-      LeakSupp
+
+      FreeSupp,      // Invalid or mismatching free
+      OverlapSupp,   // Overlapping blocks in memcpy(), strcpy(), etc
+      LeakSupp,      // Something to be suppressed in a leak check.
+      MempoolSupp,   // Memory pool suppression.
    } 
    MAC_SuppKind;
 
@@ -100,7 +94,8 @@ typedef
           ParamErr, UserErr,  /* behaves like an anonymous ParamErr */
           FreeErr, FreeMismatchErr,
           OverlapErr,
-          LeakErr
+          LeakErr,
+          IllegalMempoolErr,
    }
    MAC_ErrorKind;
 
@@ -111,15 +106,11 @@ typedef
 
 /* Extra context for memory errors */
 typedef
-   struct {
-      /* AddrErr */
-      AxsKind axskind;
-      /* AddrErr, ValueErr */
-      Int size;
-      /* AddrErr, FreeErr, FreeMismatchErr, ParamErr, UserErr */
-      AddrInfo addrinfo;
-      /* ParamErr, UserErr, CoreMemErr */
-      Bool isWrite;
+   struct {                // Used by:
+      AxsKind axskind;     //   AddrErr
+      Int size;            //   AddrErr, ValueErr
+      AddrInfo addrinfo;   //   {Addr,Free,FreeMismatch,Param,User}Err
+      Bool isWrite;        //   ParamErr, UserErr, CoreMemErr
    }
    MAC_Error;
 
@@ -146,15 +137,27 @@ typedef
 typedef
    struct _MAC_Chunk {
       struct _MAC_Chunk* next;
-      Addr          data;           /* ptr to actual block              */
-      UInt          size : 30;      /* size requested                   */
-      MAC_AllocKind allockind : 2;  /* which wrapper did the allocation */
-      ExeContext*   where;          /* where it was allocated           */
+      Addr          data;           // ptr to actual block
+      UInt          size : 30;      // size requested
+      MAC_AllocKind allockind : 2;  // which wrapper did the allocation
+      ExeContext*   where;          // where it was allocated
    }
    MAC_Chunk;
 
+/* Memory pool.  Nb: first two fields must match core's VgHashNode. */
+typedef
+   struct _MAC_Mempool {
+      struct _MAC_Mempool* next;
+      Addr          pool;           // pool identifier
+      UInt          rzB;            // pool red-zone size
+      Bool          is_zeroed;      // allocations from this pool are zeroed
+      VgHashTable   chunks;         // chunks associated with this pool
+   }
+   MAC_Mempool;
+
+
 /*------------------------------------------------------------*/
-/*--- Profiling of skins and memory events                 ---*/
+/*--- Profiling of tools and memory events                 ---*/
 /*------------------------------------------------------------*/
 
 typedef 
@@ -163,7 +166,7 @@ typedef
       VgpSetMem,
       VgpESPAdj
    } 
-   VgpSkinCC;
+   VgpToolCC;
 
 /* Define to collect detailed performance info. */
 /* #define MAC_PROFILE_MEMORY */
@@ -270,7 +273,10 @@ extern void MAC_(print_common_debug_usage)       ( void );
 /* For tracking malloc'd blocks */
 extern VgHashTable MAC_(malloc_list);
 
-/* Function pointers for the two skins to track interesting events. */
+/* For tracking memory pools. */
+extern VgHashTable MAC_(mempool_list);
+
+/* Function pointers for the two tools to track interesting events. */
 extern void (*MAC_(new_mem_heap)) ( Addr a, UInt len, Bool is_inited );
 extern void (*MAC_(ban_mem_heap)) ( Addr a, UInt len );
 extern void (*MAC_(die_mem_heap)) ( Addr a, UInt len );
@@ -298,10 +304,16 @@ extern void MAC_(clear_MAC_Error)          ( MAC_Error* err_extra );
 
 extern Bool MAC_(shared_recognised_suppression) ( Char* name, Supp* su );
 
-extern void MAC_(new_block) ( Addr p, UInt size, UInt rzB,
-                              Bool is_zeroed, MAC_AllocKind kind );
+extern void* MAC_(new_block) ( Addr p, UInt size, UInt align, UInt rzB,
+                               Bool is_zeroed, MAC_AllocKind kind,
+                               VgHashTable table);
 extern void MAC_(handle_free) ( Addr p, UInt rzB, MAC_AllocKind kind );
 
+extern void MAC_(create_mempool)(Addr pool, UInt rzB, Bool is_zeroed);
+extern void MAC_(destroy_mempool)(Addr pool);
+extern void MAC_(mempool_alloc)(Addr pool, Addr addr, UInt size);
+extern void MAC_(mempool_free)(Addr pool, Addr addr);
+
 extern void MAC_(record_address_error)     ( ThreadId tid, Addr a,
                                              Int size, Bool isWrite );
 extern void MAC_(record_core_mem_error)    ( ThreadId tid, Bool isWrite,
@@ -312,10 +324,11 @@ extern void MAC_(record_jump_error)        ( ThreadId tid, Addr a );
 extern void MAC_(record_free_error)        ( ThreadId tid, Addr a );
 extern void MAC_(record_freemismatch_error)( ThreadId tid, Addr a );
 extern void MAC_(record_overlap_error)     ( Char* function, OverlapExtra* oe );
+extern void MAC_(record_illegal_mempool_error) ( ThreadId tid, Addr pool );
 
 extern void MAC_(pp_shared_SkinError)      ( Error* err);
 
-extern MAC_Chunk* MAC_(first_matching_freed_MAC_Chunk)( Bool (*p)(MAC_Chunk*) );
+extern MAC_Chunk* MAC_(first_matching_freed_MAC_Chunk)( Bool (*p)(MAC_Chunk*, void*), void* d );
 
 extern void MAC_(common_pre_clo_init) ( void );
 extern void MAC_(common_fini)         ( void (*leak_check)(void) );
@@ -334,18 +347,18 @@ extern void MAC_(do_detect_memory_leaks) (
           Bool is_valid_address   ( Addr )
        );
 
-extern __attribute__((regparm(1))) void MAC_(new_mem_stack_4)  ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(die_mem_stack_4)  ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(new_mem_stack_8)  ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(die_mem_stack_8)  ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(new_mem_stack_12) ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(die_mem_stack_12) ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(new_mem_stack_16) ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(die_mem_stack_16) ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(new_mem_stack_32) ( Addr old_ESP );
-extern __attribute__((regparm(1))) void MAC_(die_mem_stack_32) ( Addr old_ESP );
-extern                             void MAC_(die_mem_stack) ( Addr a, UInt len);
-extern                             void MAC_(new_mem_stack) ( Addr a, UInt len);
+extern REGPARM(1) void MAC_(new_mem_stack_4)  ( Addr old_ESP );
+extern REGPARM(1) void MAC_(die_mem_stack_4)  ( Addr old_ESP );
+extern REGPARM(1) void MAC_(new_mem_stack_8)  ( Addr old_ESP );
+extern REGPARM(1) void MAC_(die_mem_stack_8)  ( Addr old_ESP );
+extern REGPARM(1) void MAC_(new_mem_stack_12) ( Addr old_ESP );
+extern REGPARM(1) void MAC_(die_mem_stack_12) ( Addr old_ESP );
+extern REGPARM(1) void MAC_(new_mem_stack_16) ( Addr old_ESP );
+extern REGPARM(1) void MAC_(die_mem_stack_16) ( Addr old_ESP );
+extern REGPARM(1) void MAC_(new_mem_stack_32) ( Addr old_ESP );
+extern REGPARM(1) void MAC_(die_mem_stack_32) ( Addr old_ESP );
+extern            void MAC_(die_mem_stack) ( Addr a, UInt len);
+extern            void MAC_(new_mem_stack) ( Addr a, UInt len);
 
 
 /*------------------------------------------------------------*/
@@ -364,7 +377,7 @@ extern                             void MAC_(new_mem_stack) ( Addr a, UInt len);
                             ALIGNED8_NEW,  ALIGNED8_DIE,                      \
                             UNALIGNED_NEW, UNALIGNED_DIE)                     \
                                                                               \
-void __attribute__((regparm(1))) MAC_(new_mem_stack_4)(Addr new_ESP)          \
+void REGPARM(1) MAC_(new_mem_stack_4)(Addr new_ESP)                           \
 {                                                                             \
    PROF_EVENT(110);                                                           \
    if (IS_ALIGNED4_ADDR(new_ESP)) {                                           \
@@ -374,7 +387,7 @@ void __attribute__((regparm(1))) MAC_(new_mem_stack_4)(Addr new_ESP)          \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(die_mem_stack_4)(Addr new_ESP)          \
+void REGPARM(1) MAC_(die_mem_stack_4)(Addr new_ESP)                           \
 {                                                                             \
    PROF_EVENT(120);                                                           \
    if (IS_ALIGNED4_ADDR(new_ESP)) {                                           \
@@ -384,7 +397,7 @@ void __attribute__((regparm(1))) MAC_(die_mem_stack_4)(Addr new_ESP)          \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(new_mem_stack_8)(Addr new_ESP)          \
+void REGPARM(1) MAC_(new_mem_stack_8)(Addr new_ESP)                           \
 {                                                                             \
    PROF_EVENT(111);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
@@ -397,7 +410,7 @@ void __attribute__((regparm(1))) MAC_(new_mem_stack_8)(Addr new_ESP)          \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(die_mem_stack_8)(Addr new_ESP)          \
+void REGPARM(1) MAC_(die_mem_stack_8)(Addr new_ESP)                           \
 {                                                                             \
    PROF_EVENT(121);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
@@ -410,7 +423,7 @@ void __attribute__((regparm(1))) MAC_(die_mem_stack_8)(Addr new_ESP)          \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(new_mem_stack_12)(Addr new_ESP)         \
+void REGPARM(1) MAC_(new_mem_stack_12)(Addr new_ESP)                          \
 {                                                                             \
    PROF_EVENT(112);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
@@ -424,7 +437,7 @@ void __attribute__((regparm(1))) MAC_(new_mem_stack_12)(Addr new_ESP)         \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(die_mem_stack_12)(Addr new_ESP)         \
+void REGPARM(1) MAC_(die_mem_stack_12)(Addr new_ESP)                          \
 {                                                                             \
    PROF_EVENT(122);                                                           \
    /* Note the -12 in the test */                                             \
@@ -439,7 +452,7 @@ void __attribute__((regparm(1))) MAC_(die_mem_stack_12)(Addr new_ESP)         \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(new_mem_stack_16)(Addr new_ESP)         \
+void REGPARM(1) MAC_(new_mem_stack_16)(Addr new_ESP)                          \
 {                                                                             \
    PROF_EVENT(113);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
@@ -454,7 +467,7 @@ void __attribute__((regparm(1))) MAC_(new_mem_stack_16)(Addr new_ESP)         \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(die_mem_stack_16)(Addr new_ESP)         \
+void REGPARM(1) MAC_(die_mem_stack_16)(Addr new_ESP)                          \
 {                                                                             \
    PROF_EVENT(123);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
@@ -469,7 +482,7 @@ void __attribute__((regparm(1))) MAC_(die_mem_stack_16)(Addr new_ESP)         \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(new_mem_stack_32)(Addr new_ESP)         \
+void REGPARM(1) MAC_(new_mem_stack_32)(Addr new_ESP)                          \
 {                                                                             \
    PROF_EVENT(114);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
@@ -488,7 +501,7 @@ void __attribute__((regparm(1))) MAC_(new_mem_stack_32)(Addr new_ESP)         \
    }                                                                          \
 }                                                                             \
                                                                               \
-void __attribute__((regparm(1))) MAC_(die_mem_stack_32)(Addr new_ESP)         \
+void REGPARM(1) MAC_(die_mem_stack_32)(Addr new_ESP)                          \
 {                                                                             \
    PROF_EVENT(124);                                                           \
    if (IS_ALIGNED8_ADDR(new_ESP)) {                                           \
