@@ -149,11 +149,6 @@ static SecMap* alloc_secondary_map ( __attribute__ ((unused))
    PROF_EVENT(10);
 
    /* Mark all bytes as invalid access and invalid value. */
-
-   /* It just happens that a SecMap occupies exactly 18 pages --
-      although this isn't important, so the following assert is
-      spurious. */
-   sk_assert(0 == (sizeof(SecMap) % VKI_BYTES_PER_PAGE));
    map = (SecMap *)VG_(shadow_alloc)(sizeof(SecMap));
 
    for (i = 0; i < 8192; i++)
@@ -767,7 +762,7 @@ static __inline__ UInt shiftRight16 ( UInt x )
    Under all other circumstances, it defers to the relevant _SLOWLY
    function, which can handle all situations.
 */
-__attribute__ ((regparm(1)))
+REGPARM(1)
 UInt MC_(helperc_LOADV4) ( Addr a )
 {
 #  ifdef VG_DEBUG_MEMORY
@@ -792,7 +787,7 @@ UInt MC_(helperc_LOADV4) ( Addr a )
 #  endif
 }
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 void MC_(helperc_STOREV4) ( Addr a, UInt vbytes )
 {
 #  ifdef VG_DEBUG_MEMORY
@@ -817,7 +812,7 @@ void MC_(helperc_STOREV4) ( Addr a, UInt vbytes )
 #  endif
 }
 
-__attribute__ ((regparm(1)))
+REGPARM(1)
 UInt MC_(helperc_LOADV2) ( Addr a )
 {
 #  ifdef VG_DEBUG_MEMORY
@@ -840,7 +835,7 @@ UInt MC_(helperc_LOADV2) ( Addr a )
 #  endif
 }
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 void MC_(helperc_STOREV2) ( Addr a, UInt vbytes )
 {
 #  ifdef VG_DEBUG_MEMORY
@@ -861,7 +856,7 @@ void MC_(helperc_STOREV2) ( Addr a, UInt vbytes )
 #  endif
 }
 
-__attribute__ ((regparm(1)))
+REGPARM(1)
 UInt MC_(helperc_LOADV1) ( Addr a )
 {
 #  ifdef VG_DEBUG_MEMORY
@@ -884,7 +879,7 @@ UInt MC_(helperc_LOADV1) ( Addr a )
 #  endif
 }
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 void MC_(helperc_STOREV1) ( Addr a, UInt vbytes )
 {
 #  ifdef VG_DEBUG_MEMORY
@@ -1112,7 +1107,7 @@ void MC_(helperc_value_check4_fail) ( void )
    FPU load and store checks, called from generated code.
    ------------------------------------------------------------------ */
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 void MC_(fpu_read_check) ( Addr addr, Int size )
 {
    /* Ensure the read area is both addressible and valid (ie,
@@ -1204,7 +1199,7 @@ void MC_(fpu_read_check) ( Addr addr, Int size )
 }
 
 
-__attribute__ ((regparm(2)))
+REGPARM(2)
 void MC_(fpu_write_check) ( Addr addr, Int size )
 {
    /* Ensure the written area is addressible, and moan if otherwise.
@@ -1457,9 +1452,9 @@ Bool mc_is_valid_address ( Addr a )
 }
 
 
-/* Leak detector for this skin.  We don't actually do anything, merely
+/* Leak detector for this tool.  We don't actually do anything, merely
    run the generic leak detector with suitable parameters for this
-   skin. */
+   tool. */
 void MC_(detect_memory_leaks) ( void )
 {
    MAC_(do_detect_memory_leaks) ( mc_is_valid_64k_chunk, mc_is_valid_address );
@@ -1601,16 +1596,8 @@ Bool  MC_(clo_cleanup)                = True;
 
 Bool SK_(process_cmd_line_option)(Char* arg)
 {
-   if (VG_CLO_STREQ(arg, "--avoid-strlen-errors=yes"))
-      MC_(clo_avoid_strlen_errors) = True;
-   else if (VG_CLO_STREQ(arg, "--avoid-strlen-errors=no"))
-      MC_(clo_avoid_strlen_errors) = False;
-
-   else if (VG_CLO_STREQ(arg, "--cleanup=yes"))
-      MC_(clo_cleanup) = True;
-   else if (VG_CLO_STREQ(arg, "--cleanup=no"))
-      MC_(clo_cleanup) = False;
-
+        VG_BOOL_CLO("--avoid-strlen-errors", MC_(clo_avoid_strlen_errors))
+   else VG_BOOL_CLO("--cleanup",             MC_(clo_cleanup))
    else
       return MAC_(process_common_cmd_line_option)(arg);
 
@@ -1644,7 +1631,7 @@ void SK_(pre_clo_init)(void)
    VG_(details_version)         (NULL);
    VG_(details_description)     ("a memory error detector");
    VG_(details_copyright_author)(
-      "Copyright (C) 2002-2004, and GNU GPL'd, by Julian Seward.");
+      "Copyright (C) 2002-2004, and GNU GPL'd, by Julian Seward et al.");
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
    VG_(details_avg_translation_sizeB) ( 228 );
 
