@@ -1,4 +1,4 @@
-/* -*- c-basic-offset: 3 -*- */
+
 /*--------------------------------------------------------------------*/
 /*--- Instrument UCode to perform memory checking operations.      ---*/
 /*---                                               mc_translate.c ---*/
@@ -8,7 +8,7 @@
    This file is part of MemCheck, a heavyweight Valgrind tool for
    detecting memory errors.
 
-   Copyright (C) 2000-2003 Julian Seward 
+   Copyright (C) 2000-2004 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -36,8 +36,8 @@
    ------------------------------------------------------------------ */
 
 /* Compare this with the restrictions on core instructions in
-   vg_translate.c:VG_(saneUInstr)().  Everything general said there applies
-   here too.
+   vg_translate.c:is_sane_UInstr().  Everything general said there
+   applies here too.
 */
 Bool SK_(sane_XUInstr)(Bool beforeRA, Bool beforeLiveness, UInstr* u)
 {
@@ -57,7 +57,6 @@ Bool SK_(sane_XUInstr)(Bool beforeRA, Bool beforeLiveness, UInstr* u)
 #  define Ls1 (u->tag1 == Lit16)
 #  define Ls3 (u->tag3 == Lit16)
 #  define TRL1 (TR1 || L1)
-#  define TRA1 (TR1 || A1)
 #  define N2  (u->tag2 == NoValue)
 #  define N3  (u->tag3 == NoValue)
 #  define COND0    (u->cond         == 0)
@@ -89,7 +88,7 @@ Bool SK_(sane_XUInstr)(Bool beforeRA, Bool beforeLiveness, UInstr* u)
    case GETVF: 
    case PUTVF:  return LIT0 && SZ0 && CC0 &&  TR1 &&  N2 &&  N3 && XOTHER;
    case TESTV: 
-   case SETV:   return LIT0 && SZj && CC0 && TRA1 &&  N2 &&  N3 && XOTHER;
+   case SETV:   return LIT0 && SZj && CC0 &&  TR1 &&  N2 &&  N3 && XOTHER;
    case TAG1:   return LIT0 && SZ0 && CC0 &&  TR1 &&  N2 && Ls3 && XOTHER;
    case TAG2:   return LIT0 && SZ0 && CC0 &&  TR1 && TR2 && Ls3 && XOTHER;
    default:
@@ -111,7 +110,6 @@ Bool SK_(sane_XUInstr)(Bool beforeRA, Bool beforeLiveness, UInstr* u)
 #  undef Ls1
 #  undef Ls3
 #  undef TRL1
-#  undef TRA1
 #  undef N2
 #  undef N3
 #  undef COND0
@@ -575,7 +573,7 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
             still addressible.  Hence the optionalisation of the V
             check.  15 Dec 02: optionalisation removed, since it no
             longer makes much sense given we also have an addrcheck
-            skin.
+            tool.
 
             The LOADV/STOREV does an addressibility check for the
             address. */
@@ -718,9 +716,9 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
                Noticed by Kevin Ryde <user42@zip.com.au>
             */
 	    /* sk_assert(u_in->flags_w != FlagsEmpty); */
-            qs = getOperandShadow(cb, u_in->size, u_in->tag1, u_in->val1);
+            qs = getOperandShadow(cb, 1, u_in->tag1, u_in->val1);
             /* We can safely modify qs; cast it to 0-size. */
-            create_PCast(cb, u_in->size, 0, qs);
+            create_PCast(cb, 1, 0, qs);
             qd = SHADOW(u_in->val2);
             create_PCast(cb, u_in->size, 0, qd);
             /* qs is cast-to-0(shift count#), and qd is cast-to-0(value#). */
@@ -1075,22 +1073,45 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
             break;
          }
 
+         case MMX2a1_MemRd: {
+            Int t_size = INVALID_TEMPREG;
+
+            sk_assert(u_in->size == 8);
+
+            sk_assert(u_in->tag3 == TempReg);
+            uInstr1(cb, TESTV, 4, TempReg, SHADOW(u_in->val3));
+            uInstr1(cb, SETV,  4, TempReg, SHADOW(u_in->val3));
+
+            t_size = newTemp(cb);
+            uInstr2(cb, MOV,   4, Literal, 0, TempReg, t_size);
+            uLiteral(cb, u_in->size);
+            uInstr2(cb, CCALL, 0, TempReg, u_in->val3, TempReg, t_size);
+            uCCall(cb, (Addr) & MC_(fpu_read_check), 2, 2, False);
+            
+            VG_(copy_UInstr)(cb, u_in);
+            break;
+         }
+
 	 /* SSE ins referencing scalar integer registers */
-	 case SSE3g_RegWr:
+         case SSE2g_RegWr:
+         case SSE2g1_RegWr:
+         case SSE2e1_RegRd:
+         case SSE3g_RegWr:
          case SSE3e_RegRd:
          case SSE3e_RegWr: 
          case SSE3g1_RegWr:
          case SSE3e1_RegRd:
             sk_assert(u_in->tag3 == TempReg);
 
-            if (u_in->opcode == SSE3e1_RegRd) {
+            if (u_in->opcode == SSE2e1_RegRd || u_in->opcode == SSE3e1_RegRd) {
                sk_assert(u_in->size == 2);
             } else {
                sk_assert(u_in->size == 4);
             }
 
             /* Is it a read ?  Better check the V bits right now. */
-            if ( u_in->opcode == SSE3e_RegRd
+            if ( u_in->opcode == SSE2e1_RegRd
+                 || u_in->opcode == SSE3e_RegRd
                  || u_in->opcode == SSE3e1_RegRd )
 	       uInstr1(cb, TESTV, u_in->size, 
                            TempReg, SHADOW(u_in->val3));
@@ -1108,16 +1129,18 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
          case SSE3a_MemWr:
          case SSE2a_MemWr:
          case SSE2a_MemRd:
-         case SSE3a1_MemRd: { 
+         case SSE3a1_MemRd:
+         case SSE2a1_MemRd: { 
             Bool is_load;
             Int t_size;
 
-            sk_assert(u_in->size == 4 
-                      || u_in->size == 8 || u_in->size == 16);
+            sk_assert(u_in->size == 4 || u_in->size == 8
+                      || u_in->size == 16 || u_in->size == 512);
 
             t_size = INVALID_TEMPREG;
             is_load = u_in->opcode==SSE2a_MemRd 
                       || u_in->opcode==SSE3a_MemRd
+                      || u_in->opcode==SSE2a1_MemRd
                       || u_in->opcode==SSE3a1_MemRd;
 
             sk_assert(u_in->tag3 == TempReg);
@@ -1153,12 +1176,26 @@ static UCodeBlock* memcheck_instrument ( UCodeBlock* cb_in )
             VG_(copy_UInstr)(cb, u_in);
 	    break;
          }
-
-         /* For FPU, MMX and SSE insns not referencing memory, just
-            copy thru. */
+         
+         /* For MMX and SSE insns not referencing memory, just
+            make sure the eflags are defined if the instruction
+            read them, and make them defined it it writes them. */
          case SSE5: case SSE4: case SSE3:
          case MMX1: case MMX2: case MMX3:
-         case FPU: 
+         case FPU:
+            if (u_in->flags_r != FlagsEmpty) {
+               qt = create_GETVF(cb, 0);
+               uInstr1(cb, TESTV, 0, TempReg, qt);
+               /* qt should never be referred to again.  Nevertheless
+                  ... */
+               uInstr1(cb, SETV, 0, TempReg, qt);
+            }
+            if (u_in->flags_w != FlagsEmpty) {
+               qd = newTemp(cb);
+               uInstr2(cb, MOV, 4, Literal, 0, TempReg, qd);
+               uLiteral(cb, 0);
+               create_PUTVF(cb, 0, qd);
+            }
             VG_(copy_UInstr)(cb, u_in);
             break;
 
@@ -1491,7 +1528,7 @@ static void vg_propagate_definedness ( UCodeBlock* cb )
                case Tag_ImproveAND1_TQ:
                   sz = 1; goto do_ImproveAND;
                do_ImproveAND:
-                  /* Implements Q = T OR Q.  So if Q is entirely defined,
+                  /* Implements Q = T AND Q.  So if Q is entirely defined,
                      ie all 0s, we get MOV T, Q. */
 		  if (def[u->val2] <= 4) {
                      sk_assert(def[u->val2] == sz);
